try:
    import PySide2  # this will force pyqtgraph to use PySide instead of PyQt4/5
except ImportError:
    pass
from pyqtgraph.Qt import QtWidgets, QtCore, QtGui
import numpy as np

if not hasattr(QtCore, 'Slot'):
    QtCore.Slot = QtCore.pyqtSlot


class Table(QtGui.QDialog):

    def __init__(self, data=[], model=None, as_dialog=True, **kwargs):
        super().__init__(**kwargs)
        self.title = 'Edit song definitions'

        self.data = data
        self.model = model
        self.as_dialog = as_dialog
        self.cancelled = False
        self.initUI(self.as_dialog)

    def initUI(self, as_dialog=True):
        self.setWindowTitle(self.title)
        # self.setGeometry(self.left, self.top, self.width, self.height)

        self.createTable()
        self.createButtons()

        # Add box layout, add table to box layout and add box layout to widget
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.table)

        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.addWidget(self.add_button)
        self.button_layout.addWidget(self.delete_button)
        self.button_layout.addWidget(self.load_button)
        self.button_layout.addWidget(self.save_button)
        self.layout.addLayout(self.button_layout)

        if as_dialog:
            buttons = QtGui.QDialogButtonBox(
                QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel,
                QtCore.Qt.Horizontal, self)
            buttons.accepted.connect(self.accept)
            buttons.rejected.connect(self.reject)
            self.layout.addWidget(buttons)

        self.setLayout(self.layout)

    def createTable(self):
        # Create table
        self.table = QtWidgets.QTableWidget()
        self.table.setRowCount(len(self.data))
        if len(self.data) > 0:
            self.table.setColumnCount(len(self.data[0]))
        else:
            self.table.setColumnCount(2)
        self.table.setHorizontalHeaderLabels(['Name', 'Category'])

        for row_index, row_data in enumerate(self.data):
            self._make_row(row_index, row_data, editable_categories=False)
        self.table.move(0, 0)

    def _make_row(self, row_index=None,
                  row_data=['', 'segment'],
                  row_items=["segment", "event"],
                  editable_categories=True):
        """[summary]

        Args:
            row_index ([type], optional): If None or omitted, will add row at end of table.
                                          Defaults to None.
            row_data (list, optional): [description]. Defaults to ['', 'segment'].
            row_items (list, optional): Dropdown items for second columns. Defaults to ["segment", "event"].
        """
        if row_index is None:
            row_index = self.table.rowCount()
            self.table.insertRow(row_index)

        item = QtWidgets.QTableWidgetItem(row_data[0])
        item.original_text = row_data[0]
        self.table.setItem(row_index, 0, item)

        if editable_categories:
            cb = QtWidgets.QComboBox()
            cb.addItems(row_items)
            if row_data[1] in row_items:
                cb.setCurrentText(row_data[1])
                cb.original_text = row_data[1]
            else:
                cb.original_text = ''

            self.table.setCellWidget(row_index, 1, cb)
        else:
            item = QtWidgets.QTableWidgetItem(row_data[1])
            item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEnabled & ~QtCore.Qt.ItemIsEditable)
            item.original_text = row_data[1]
            self.table.setItem(row_index, 1, item)

    def get_cell_data(self, row: int, col: int):
        if self.table.item(row, col) is not None:
            return [self.table.item(row, col).text(), self.table.item(row, col).original_text]
        elif self.table.cellWidget(row, col) is not None:
            return [self.table.cellWidget(row, col).currentText(), self.table.cellWidget(row, col).original_text]
        else:
            return None

    def get_table_data(self):
        rows = self.table.rowCount()
        cols = self.table.columnCount()
        data = []
        for row in range(rows):
            row_data = []
            for col in range(cols):
                row_data.append(self.get_cell_data(row, col))
            data.append(row_data)
        return data

    def createButtons(self):
        self.add_button = QtWidgets.QPushButton('New', self)
        self.add_button.clicked.connect(self.add_event)

        self.delete_button = QtWidgets.QPushButton('Delete', self)
        self.delete_button.clicked.connect(self.delete_event)

        self.load_button = QtWidgets.QPushButton(self.tr("&Load"), self)
        self.load_button.clicked.connect(self.load)

        self.save_button = QtWidgets.QPushButton(self.tr("&Save"), self)
        self.save_button.clicked.connect(self.save)

    @QtCore.Slot()
    def cancel(self):
        self.cancelled = True

    @QtCore.Slot()
    def add_event(self):
        self.data.append(['', 'segment'])
        self._make_row()

    @QtCore.Slot()
    def delete_event(self):
        del self.data[self.table.currentRow()]
        self.table.removeRow(self.table.currentRow())

    @QtCore.Slot()
    def load(self):
        filename = self.model._get_filename(suffix="_definitions.csv")
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            None, caption="Select definitions file", dir=filename, filter=self.tr("Any File (*_definitions.csv)")
        )
        data = np.loadtxt(filename, dtype=str, delimiter=",")
        # delete existing data
        self.data = []
        while self.table.rowCount() > 0:
            self.table.removeRow(self.table.currentRow())

        # replace with loaded data
        for d in data:
            self.data.append(d)
            self._make_row(row_data=d)

    @QtCore.Slot()
    def save(self):
        savefilename = self.model._get_filename(suffix="_definitions.csv")

        savefilename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, caption="Select file for saving", dir=savefilename, filter=self.tr("Any File (*_definitions.csv)")
        )
        data = self.get_table_data()
        data = [[d[0][0], d[1][0]] for d in data]
        np.savetxt(savefilename, data, delimiter=",", fmt="%s")



import h5py
import flammkuchen
import numpy as np
import pandas as pd
import scipy.io
import logging
from typing import Optional
from .. import (io,
                annot)


def fill_gaps(sine_pred, gap_dur=100):
    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets<offsets[-1]]
        offsets = offsets[offsets>onsets[0]]
        durations = offsets - onsets
        for idx, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
            if idx>0 and offsets[idx-1]>onsets[idx]-gap_dur:
                sine_pred[offsets[idx-1]:onsets[idx]+1] = 1
    return sine_pred


def remove_short(sine_pred, min_len=100):
    # remove too short sine songs
    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets<offsets[-1]]
        offsets = offsets[offsets>onsets[0]]
        durations = offsets - onsets
        for cnt, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
            if duration<min_len:
                sine_pred[onset:offset+1] = 0
    return sine_pred


@io.register_provider
class DAS(io.BaseProvider):

    KIND = 'annotations'
    NAME = 'das'
    SUFFIXES = ['_song.h5', '_vibration.h5', '_pulse.h5', '_sine.h5', '_dss.h5', '_das.h5']

    def load(self, filename: Optional[str] = None):
        """Load output produced by DeepAudioSegmenter wrapper."""
        if filename is None:
            filename = self.path

        onsets = []
        offsets = []
        names = []

        res = flammkuchen.load(filename)

        if 'event_indices' in res or 'segment_labels' in res:  # load old style format
            logging.info('   Converting legacy DAS format...')
            res['event_seconds'] = np.zeros((0,))
            res['event_sequence'] = []
            for event_name, event_idx in zip(res['event_names'], res['event_indices']):
                res['event_seconds'] = np.append(res['event_seconds'], np.array(event_idx) / res['samplerate_Hz'])
                res['event_sequence'] = np.append(res['event_sequence'], [event_name for _ in range(len(event_idx))])

            res['segment_onsets_seconds'] = np.zeros((0,))
            res['segment_offsets_seconds'] = np.zeros((0,))
            res['segment_sequence'] = []
            for segment_name, segment_labels in zip(res['segment_names'], res['segment_labels']):
                if 'sine' in segment_name:
                    logging.info(f'   Postprocessing {segment_name}')
                    segment_labels = fill_gaps(segment_labels, gap_dur=0.02 * res['samplerate_Hz'])
                    segment_labels = remove_short(segment_labels, min_len=0.02 * res['samplerate_Hz'])

                # detect on and offset from binary labels
                segment_onset_idx = np.where(np.diff(segment_labels.astype(np.float), prepend=0)==1)[0].astype(np.float)
                segment_offset_idx = np.where(np.diff(segment_labels.astype(np.float), append=0)==-1)[0].astype(np.float)

                res['segment_onsets_seconds'] = np.append(res['segment_onsets_seconds'], segment_onset_idx / res['samplerate_Hz'])
                res['segment_offsets_seconds'] = np.append(res['segment_offsets_seconds'], segment_offset_idx / res['samplerate_Hz'])
                res['segment_sequence'] = np.append(res['segment_sequence'], [segment_name for _ in range(len(segment_onset_idx))])

        for event_seconds, event_names in zip(res['event_seconds'], res['event_sequence']):
            onsets.append(event_seconds)
            offsets.append(event_seconds)
            names.append(event_names)

        if 'event_names' in res:  # ensure empty event types are initialized
            for name in res['event_names']:
                if name not in names:
                    names.append(name)
                    onsets.append(np.nan)
                    offsets.append(np.nan)

        for segment_onsets, segment_offsets, segment_names in zip(res['segment_onsets_seconds'], res['segment_offsets_seconds'], res['segment_sequence']):
            onsets.append(segment_onsets)
            offsets.append(segment_offsets)
            names.append(segment_names)

        if 'segment_names' in res:  # ensure empty segment types are initialized
            for name in res['segment_names']:
                if name not in names:
                    names.append(name)
                    onsets.append(np.nan)
                    offsets.append(0)

        et = annot.Events.from_lists(names=names, start_seconds=onsets, stop_seconds=offsets)

        return et, et.categories


@io.register_provider
class FlySongSegmenter(io.BaseProvider):

    KIND = 'annotations'
    NAME = 'FlySongSegmenter'
    SUFFIXES = ['_song.mat']

    def load(self, filename: Optional[str] = None):
        """Load output produced by FlySongSegmenter."""
        if filename is None:
            filename = self.path

        res = dict()
        try:
            d = scipy.io.loadmat(filename)
            res['pulse_times_samples'] = d['pInf'][0, 0]['wc']
            res['pulse_labels'] = d['pInf'][0, 0]['pulseLabels']
            res['song_mask'] = d['bInf'][0, 0]['Mask'].T  # 0 - silence, 1 - pulse, 2 - sine
        except NotImplementedError:
            with h5py.File(filename, 'r') as f:
                res['pulse_times_samples'] = f['pInf/wc'][:].T
                res['pulse_labels'] = f['pInf/pulseLabels'][:].T
                res['song_mask'] = f['bInf/Mask'][:]

        sine_song = res['song_mask'] == 2
        fs = 10_000  # Hz

        res['event_names'] = ['song_pulse_any_fss', 'song_pulse_slow_fss', 'song_pulse_fast_fss', 'sine_fss']
        res['event_categories'] = ['event', 'event', 'event', 'segment']
        res['event_indices'] = [res['pulse_times_samples'],
                                res['pulse_times_samples'][res['pulse_labels'] == 1],
                                res['pulse_times_samples'][res['pulse_labels'] == 0],
                                np.where(sine_song == 2)[0]]
        # extract event_seconds
        event_seconds = {'song_pulse_any_fss': res['pulse_times_samples'] / fs,
                        'song_pulse_slow_fss': res['pulse_times_samples'][res['pulse_labels'] == 1] / fs,
                        'song_pulse_fast_fss': res['pulse_times_samples'][res['pulse_labels'] == 0] / fs,
                        'sine_fss': np.where(sine_song == 2)[0] / fs}
        # event_categories
        event_categories = {}
        for cat, typ in zip(event_seconds.keys(), ['event', 'event', 'event', 'segment']):
            event_categories[cat] = typ
        return event_seconds, event_categories


import logging
import numpy as np
import xarray as xr
import pandas as pd

def detect_events(ds):
    """Transform ds.song_events into dict of event (on/offset) times.
    Args:
        ds ([xarray.Dataset]): dataset with song_events

    Returns:
        dict: with event times or segment on/offsets.
    """
    event_times = dict()
    ds.song_events.data = ds.song_events.data.astype(np.float)  # make sure this is non-bool so diff works
    event_names = ds.song_events.event_types.data
    event_categories = ds.song_events.event_categories.data
    logging.info('Extracting event times from song_events:')
    for event_idx, (event_name, event_category) in enumerate(zip(event_names, event_categories)):
        logging.info(f'   {event_name}')
        if event_category == 'event':
            event_times[event_name] = ds.song_events.time.where(ds.song_events[:, event_idx] == 1, drop=True).data.ravel()
        elif event_category == 'segment':
            onsets = ds.song_events.time.where(ds.song_events[:, event_idx].diff(dim='time') == 1, drop=True).data
            offsets = ds.song_events.time.where(ds.song_events[:, event_idx].diff(dim='time') == -1, drop=True).data
            if len(onsets) and len(offsets):
                # ensure onsets and offsets match
                offsets = offsets[offsets>np.min(onsets)]
                onsets = onsets[onsets<np.max(offsets)]
            if len(onsets) != len(offsets):
                logging.warning('Inconsistent segment onsets or offsets - ignoring all on- and offsets.')
                onsets = []
                offsets = []

            if not len(onsets) and not len(offsets):
                event_times[event_name] = np.zeros((0,2))
            else:
                event_times[event_name] = np.stack((onsets, offsets)).T
    return event_times


def infer_class_info_from_df(df: pd.DataFrame):
    """Based on difference between start_seconds/stop_seconds

    Args:
        df ([type]): [description]

    Returns:
        [type]: [description]
    """
    class_names, first_indices = np.unique(df['name'], return_index=True)
    class_names = list(class_names)
    class_names.insert(0, 'noise')

    # infer class type - event if start and end are the same
    class_types = ['segment']
    for first_index in first_indices:
        if df.loc[first_index]['start_seconds']==df.loc[first_index]['stop_seconds']:
            class_types.append('event')
        else:
            class_types.append('segment')
    return class_names, class_types


def infer_event_categories_from_traces(data):
    """Infer event type (event/segment) based on median interval between
    event times.

    Args:
        data ([type]): binary matrix [samples x events]
    """
    event_categories = []
    for evt in range(data.shape[1]):
        if np.any(data[:, evt]):
            dt = np.nanmedian(np.diff(np.where(data[:, evt])[0]))
        else:
            dt = 10  # fall back to events w/o annotations

        if dt > 1:
            event_categories.append('event')
        else:
            event_categories.append('segment')
    return event_categories


def infer_event_categories_from_shape(event_times):
    """Based on shape of time lists."""
    event_categories = {}
    for event_name, event_data in event_times.items():
        event_categories[event_name] = 'event' if event_data.ndim==1 else 'segment'
    return event_categories


def update_traces(ds, event_times):
    """Slightly redundant with eventtimes_to_traces but
    will add populate new events.
    """
    ## event_times to ds.song_events
    # make new song_events DataArray
    if 'song_events' in ds:
        old_values = ds.song_events.values.copy()
        attrs = ds.song_events.attrs.copy()
    else:
        old_values = np.zeros((ds.time.shape[0], 0))
        attrs = {'sampling_rate_Hz': ds.attrs['target_sampling_rate_Hz']}

    new_values = np.zeros_like(old_values, shape=(old_values.shape[0], len(event_times)))
    fs = ds.song_events.attrs['sampling_rate_Hz']
    event_categories = infer_event_categories_from_shape(event_times)
    # populate with data:
    logging.info(f'Updating:')
    for cnt, (event_name, event_data) in enumerate(event_times.items()):
        logging.info(f'   {event_name} ({event_data.shape[0]} instances')
        if event_categories[event_name] == 'event':
            event_indices = (event_data * fs).astype(np.uintp)
            event_indices = event_indices[event_indices>=0]
            event_indices = event_indices[event_indices<len(new_values)]
            new_values[event_indices, cnt] = 1
        elif event_categories[event_name] == 'segment':
            for onset, offset in event_data:
                onset_idx = (onset * fs).astype(np.uintp)
                offset_idx = (offset * fs).astype(np.uintp)
                within_bounds = onset_idx < len(new_values) and offset_idx < len(new_values)
                greater_zero = onset_idx >= 0 and offset_idx >= 0
                if within_bounds and greater_zero:
                    new_values[onset_idx:offset_idx, cnt] = 1
    logging.info(f'Done:')

    # rebuild dataset
    song_events = xr.DataArray(data=new_values,
                            dims=['time', 'event_types'],
                            coords={'time': ds.time,
                                    'event_types': list(event_times.keys()),
                                    'event_categories': (('event_types'), list(event_categories.values()))},
                            attrs=attrs)
    # delete old
    if 'song_events' in ds:
        del ds['song_events']
        del ds.coords['event_types']
        del ds.coords['event_categories']
    # add new
    ds = xr.merge((ds, song_events.to_dataset(name='song_events')))
    return ds


def eventtimes_to_traces(ds, event_times):
    """Update events in ds.song_events from dict.

    Does not add new events (events that exist in event_times but not in ds.song_events)!!

    Args:
        ds ([xarray.Dataset]): dataset with song_events
        event_times ([dict]): event times or segment on/offsets.

    Returns:
        xarray.Dataset
    """
    event_names = ds.song_events.event_types.data
    event_categories = ds.song_events.event_categories.data
    logging.info('Updating song_events from event_times:')
    for event_idx, (event_name, event_category) in enumerate(zip(event_names, event_categories)):
        logging.info(f'   {event_name}')
        ds.song_events.sel(event_types=event_name).data[:] = 0  # delete all events
        if event_category == 'event':
            # times = ds.song_events.time.sel(time=event_times[event_name][:, 0], method='nearest').data
            times = ds.song_events.time.sel(time=event_times[event_name].ravel(), method='nearest').data
            times = np.unique(times)

            # this is sloooooow
            for time in times:
                idx = np.where(ds.time==time)[0]
                ds.song_events[idx, event_idx] = 1
        elif event_category == 'segment':
            if event_times[event_name].shape[0] > 0:
                for onset, offset in zip(event_times[event_name][:, 0], event_times[event_name][:, 1]):
                    ds.song_events.sel(time=slice(onset, offset), event_types=event_name).data[:] = 1
    logging.info(f'Done.')
    return ds


def traces_to_eventtimes(traces, event_names, event_categories):
    """[summary]

    Args:
        traces ([type]): list of numpy arrays with the binary traces for each event/segment
        event_names ([type]): [description]
        event_categories ([type]): [description]

    Returns:
        [type]: [description]
    """
    assert len(traces) == len(event_names)
    assert len(traces) == len(event_categories)

    event_times = dict()

    logging.info('Extracting event times from song_events:')
    for event_idx, (event_name, event_category) in enumerate(zip(event_names, event_categories)):
        logging.info(f'   {event_name}')
        if event_category == 'event':
            event_times[event_name] = np.where(traces[event_idx] == 1)[0]
        elif event_category == 'segment':
            onsets = np.where(np.diff(traces[event_idx]) == 1)[0]
            offsets = np.where(np.diff(traces[event_idx]) == -1)[0]
            if len(onsets) and len(offsets):
                # ensure onsets and offsets match
                offsets = offsets[offsets>np.min(onsets)]
                onsets = onsets[onsets<np.max(offsets)]
            if len(onsets) != len(offsets):
                logging.warning('Inconsistent segment onsets or offsets - ignoring all on- and offsets.')
                onsets = []
                offsets = []

            if not len(onsets) and not len(offsets):
                event_times[event_name] = np.zeros((0,2))
            else:
                event_times[event_name] = np.stack((onsets, offsets)).T
    return event_times



"""Audio loader

should return:
    audio_data: np.array[time, samples]
    non_audio_data: np.array[time, samples]
    samplerate: Optional[float]
"""
# [x] daq.h5
# [x] wav, ....
# [x] npz, npy
# [ ] npy_dir


"""Tracks loaders

should return:
    x: np.array[frames, flies, body_parts, x/y]
    body_parts: List[str]
    first_tracked_frame: int
    last_tracked_frame: int
    background: np.array[width, height, pixels?]
"""

import h5py
import numpy as np
from .. import io
from typing import Optional, Sequence


def split_song_and_nonsong(data, song_channels = None, return_nonsong_channels = False):
        song = data
        nonsong = None
        if song_channels is not None:
            song = song[:, song_channels]
            if  return_nonsong_channels:
                nonsong = np.delete(data, song_channels, axis=-1)
        return song, nonsong


@io.register_provider
class Ethodrome(io.BaseProvider):

    KIND = 'audio'
    NAME = 'ethodrome h5'
    SUFFIXES = ['_daq.h5']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             **kwargs):
        """[summary]

        Args:
            filename ([type]): [description]
            song_channels (List[int], optional): Sequence of integers as indices into 'samples' datasaet.
                                                Defaults to [0,..., 15].
            return_nonsong_channels (bool, optional): will return the data not in song_channels as separate array. Defaults to False
            lazy (bool, optional): If True, will load song as dask.array, which allows lazy indexing.
                                Otherwise, will load the full recording from disk (slow). Defaults to False.

        Returns:
            [type]: [description]
        """
        if filename is None:
            filename = self.path

        if song_channels is None:  # the first 16 channels in the data are the mic recordings
            song_channels = np.arange(16)

        non_song = None
        samplerate = None
        if lazy:
            f = h5py.File(filename, mode='r', rdcc_w0=0, rdcc_nbytes=100 * (1024 ** 2), rdcc_nslots=50000)
            # convert to dask array since this allows lazily evaluated indexing...
            import dask.array as daskarray
            da = daskarray.from_array(f['samples'], chunks=(10000, 1))
            nb_channels = f['samples'].shape[1]
            song_channels = song_channels[song_channels < nb_channels]
            song = da[:, song_channels]
            if return_nonsong_channels:
                non_song_channels = list(set(list(range(nb_channels))) - set(song_channels))
                non_song = da[:, non_song_channels]
        else:
            with h5py.File(filename, 'r') as f:
                nb_channels = f['samples'].shape[1]
                song_channels = song_channels[song_channels < nb_channels]
                song = f['samples'][:, song_channels]
                if return_nonsong_channels:
                    non_song_channels = list(set(list(range(nb_channels))) - set(song_channels))
                    non_song = f['samples'][:, non_song_channels]

        return song, non_song, samplerate


@io.register_provider
class Npz(io.BaseProvider):

    KIND = 'audio'
    NAME = 'npz'
    SUFFIXES = ['.npz']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             audio_dataset: Optional[str] = None,
             **kwargs):

        if filename is None:
            filename = self.path
        if audio_dataset is None:
            audio_dataset = 'data'

        with np.load(filename) as file:
            try:
                sampling_rate = file['samplerate']
            except KeyError:
                try:
                    sampling_rate = file['samplerate_Hz']
                except KeyError:
                    sampling_rate = None
            data = file[audio_dataset]

        data = data[:, np.newaxis] if data.ndim==1 else data  # adds singleton dim for single-channel wavs

        if song_channels is None:  # the first 16 channels in the data are the mic recordings
            song_channels = np.arange(np.min((16, data.shape[1])))

        # split song and non-song channels
        song, non_song = split_song_and_nonsong(data, song_channels, return_nonsong_channels)

        return data, non_song, sampling_rate


@io.register_provider
class Npy(io.BaseProvider):

    KIND = 'audio'
    NAME = 'npy'
    SUFFIXES = ['.npy']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             **kwargs):

        if filename is None:
            filename = self.path

        data = np.load(filename)
        sampling_rate = None

        song, non_song = split_song_and_nonsong(data, song_channels, return_nonsong_channels)
        return song, non_song, sampling_rate


@io.register_provider
class Wav(io.BaseProvider):

    KIND = 'audio'
    NAME = 'wav'
    SUFFIXES = ['.wav']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             audio_dataset: Optional[str] = None):

        if filename is None:
            filename = self.path

        import scipy.io.wavfile
        sampling_rate, data = scipy.io.wavfile.read(filename)
        data = data[:, np.newaxis] if data.ndim==1 else data  # adds singleton dim for single-channel wavs

        song, non_song = split_song_and_nonsong(data, song_channels, return_nonsong_channels)
        return song, non_song, sampling_rate


@io.register_provider
class AudioFile(io.BaseProvider):

    KIND = 'audio'
    NAME = 'generic audio file'
    SUFFIXES = ['.wav', '.aif', '.mp3', '.flac']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             **kwargs):

        if filename is None:
            filename = self.path

        import soundfile
        data, sampling_rate = soundfile.read(filename)
        data = data[:, np.newaxis] if data.ndim==1 else data  # adds singleton dim for single-channel wavs

        song, non_song = split_song_and_nonsong(data, song_channels, return_nonsong_channels)
        return song, non_song, sampling_rate


@io.register_provider
class H5file(io.BaseProvider):

    KIND = 'audio'
    NAME = 'h5'
    SUFFIXES = ['.h5', '.hdf5', '.hdfs']

    def load(self, filename: Optional[str], song_channels: Optional[Sequence[int]] = None,
             return_nonsong_channels: bool = False, lazy: bool = False,
             audio_dataset: Optional[str] = None,
             **kwargs):

        if filename is None:
            filename = self.path
        if audio_dataset is None:
            audio_dataset = 'data'

        import h5py
        sampling_rate = None
        with h5py.File(filename, mode='r') as file:
            data = file[audio_dataset][:]
            try:
                sampling_rate = file.attrs['samplerate']
            except:
                pass
            try:
                sampling_rate = file['samplerate'][0]
            except:
                pass

        data = data[:, np.newaxis] if data.ndim==1 else data  # adds singleton dim for single-channel wavs
        song, non_song = split_song_and_nonsong(data, song_channels, return_nonsong_channels)
        return song, non_song, sampling_rate



import sys
import logging
import collections
from glob import glob
import os.path
from typing import Optional, Dict

import zarr
import numpy as np

import das.make_dataset as dsm
import das.npy_dir
import das.annot

import scipy.signal
import pandas as pd
import matplotlib.pyplot as plt
import scipy.io.wavfile

from .. import annot

import das.make_dataset
import das.npy_dir
import das.train
import das.predict

import xarray_behave


package_dir = xarray_behave.__path__[0]


# these should be xb.loaders!!
def data_loader_wav(filename):
    # load the recording
    fs, x = scipy.io.wavfile.read(filename)
    x = x[:, np.newaxis] if x.ndim==1 else x  # adds singleton dim for single-channel wavs
    return fs, x

def data_loader_npz(filename):
    # load the recording
    file = np.load(filename)
    fs = file['samplerate']
    x = file['data']
    x = x[:, np.newaxis] if x.ndim==1 else x  # adds singleton dim for single-channel wavs
    return fs, x


# auto register with decorator
data_loaders = {'npz': data_loader_npz, 'wav': data_loader_wav, None: None}


def load_data(filename):
    extension = os.path.splitext(filename)[1][1:]  # [1:] strips '.'
    data_loader = data_loaders[extension]
    if data_loader is not None:
        fs, x = data_loader(filename)
        return fs, x
    else:
        return None


def make(data_folder: str, store_folder: str,
         file_splits: Dict[str, float], data_splits: Dict[str, float],
         make_single_class_datasets: bool = False,
         split_in_two: bool = False,
         block_stratify: bool = False,
         block_size: int = 5,
         event_std_seconds: float = 0,
         gap_seconds: float = 0,
         delete_intermediate_store: bool = True,
         make_onset_offset_events: float = False,
         seed: Optional[float] = None):

    annotation_loader = pd.read_csv
    files_annotation = glob(data_folder + '/*_annotations.csv')

    # go through all annotation files and collect info on classes
    class_names = []
    class_types = []
    for file_annotation in files_annotation:
        print(file_annotation)
        df = annotation_loader(file_annotation)
        event_times = annot.Events.from_df(df)
        class_names.extend(event_times.names)
        class_types.extend(event_times.categories.values())

    class_names, first_indices = np.unique(class_names, return_index=True)
    class_types = list(np.array(class_types)[first_indices])
    class_names = list(class_names)

    for class_name, class_type in zip(class_names, class_types):
        print(f'found {class_name} of type {class_type}')

    class_names.insert(0, 'noise')
    class_types.insert(0, 'segment')

    if make_onset_offset_events:
        class_names.extend(['syllable_onset', 'syllable_offset'])
        class_types.extend(['event', 'event'])

    file_bases = [f[:-len('_annotations.csv')] for f in files_annotation]
    data_files = []
    for file_base in file_bases:
        if os.path.exists(file_base + '.npz'):
            data_files.append(file_base + '.npz')
        elif os.path.exists(file_base + '.wav'):
            data_files.append(file_base + '.wav')
        else:
            data_files.append(None)

    # read first data file to infer number of audio channels
    # get valid file
    dfs = [data_file for data_file in data_files if data_file is not None]
    if not len(dfs):
        logging.exception('No valid data files found.')
        raise ValueError('No valid data files found.')

    fs, x = load_data(dfs[0])
    fs = float(fs)
    nb_channels = x.shape[1]

    store = dsm.init_store(nb_channels=nb_channels,
                        nb_classes=len(class_names),
                        samplerate=fs,
                        class_names=class_names,
                        class_types=class_types,
                        make_single_class_datasets=make_single_class_datasets,
                        store_type=zarr.TempStore, store_name='store.zarr', chunk_len=1_000_000)

    # save args for reproducibility
    store.attrs['event_std_seconds'] = event_std_seconds
    store.attrs['gap_seconds'] = gap_seconds
    store.attrs['data_folder'] = data_folder
    store.attrs['store_folder'] = store_folder
    store.attrs['file_splits'] = file_splits
    store.attrs['data_splits'] = data_splits
    store.attrs['make_single_class_datasets'] = make_single_class_datasets
    store.attrs['split_in_two'] = split_in_two
    store.attrs['delete_intermediate_store'] = delete_intermediate_store

    # first split files into a train-test and val
    file_split_dict = collections.defaultdict(lambda: None)
    parts = ['train', 'val', 'test']

    if len(file_splits):
        if sum(file_splits.values())>1.0:
            raise ValueError(f'Sum of file splits > 1.0!')

        if len(file_splits)<3:
            file_splits['remainder'] = max(0, 1 - sum(file_splits.values()))

        file_splits = dsm.generate_file_splits(file_bases,
                                            splits=list(file_splits.values()),
                                            split_names=list(file_splits.keys()),
                                            seed=seed)

        for file_base in file_bases:
            for part in parts:
                if part in file_splits and file_base in file_splits[part]:
                    file_split_dict[file_base] = part
    store.attrs['file_split_dict'] = dict(file_split_dict)

    data_split_targets = []
    if len(data_splits):
        fractions = np.array(list(data_splits.values()))
        probs = fractions / np.sum(fractions)
        data_splits = {key: val for key, val in zip(data_splits.keys(), probs)}
        data_split_targets = list(data_splits.keys())
        data_splits = list(data_splits.values())
        if split_in_two:
            if 'train' in data_split_targets:
                train_idx = data_split_targets.index('train')
                data_splits[train_idx] /= 2
                data_splits.append(data_splits[train_idx])
                data_split_targets.append('train')

            if 'val' in data_split_targets:
                val_idx = data_split_targets.index('val')
                data_splits[val_idx] /= 2
                data_splits.append(data_splits[val_idx])
                data_split_targets.append('val')

            if 'test' in data_split_targets:
                test_idx = data_split_targets.index('test')
                data_splits[test_idx] /= 2
                data_splits.append(data_splits[test_idx])
                data_split_targets.append('test')

    for file_base, data_file in zip(file_bases, data_files):
        if data_file is None:
            logging.warning(f'Unknown data file for {file_base} - skipping.')
            continue
        logging.info(f'  {file_base}')

        # load the recording
        if data_file.endswith('.npz'):
            data_loader = data_loader_npz
        elif data_file.endswith('.wav'):
            data_loader = data_loader_wav
        else:
            logging.warning(f'Unknown data file {file_base} - skipping.')
            continue

        fs, x = data_loader(data_file)
        nb_samples = x.shape[0]

        # load annotations
        df = annotation_loader(file_base + '_annotations.csv')
        df = df.dropna()

        # make new events from syllable on- and offsets
        if make_onset_offset_events:
            adf = das.annot.Events.from_df(df)
            start_seconds = []
            stop_seconds = []
            for nam, cat in adf.categories.items():
                if cat == 'segment':
                    start_seconds.extend(adf.start_seconds(nam))
                    stop_seconds.extend(adf.stop_seconds(nam))

            logging.info(f'   Making syllable {len(start_seconds) + len(stop_seconds)} onset/offset events.')
            adf.add_name(name='syllable_onset', category='event')
            adf.add_name(name='syllable_offset', category='event')
            for start, stop in zip(start_seconds, stop_seconds):
                adf.add_time(name='syllable_onset', start_seconds=start)
                adf.add_time(name='syllable_offset', start_seconds=stop)

            df = adf.to_df()

        # make initial annotation matrix
        y = dsm.make_annotation_matrix(df, nb_samples, fs, class_names)

        # blur events
        # OPTIONAL but highly recommended
        if event_std_seconds > 0:
            for class_index, class_type in enumerate(class_types):
                if class_type == 'event':
                    y[:, class_index] = dsm.blur_events(y[:, class_index],
                                                        event_std_seconds=event_std_seconds,
                                                        samplerate=fs)

        # introduce small gaps between adjacent segments
        # OPTIONAL but makes boundary detection easier
        if gap_seconds > 0:
            segment_idx = np.where(np.array(class_types)=='segment')[0]
            if len(segment_idx):
                y[:, segment_idx] = dsm.make_gaps(y[:, segment_idx],
                                                  gap_seconds=gap_seconds,
                                                  samplerate=fs,
                                                  start_seconds=df['start_seconds'],
                                                  stop_seconds=df['start_seconds'])
        # all validation files
        if file_split_dict[file_base] is not None:
            name = file_split_dict[file_base]
            logging.info(f'    {name} data.')
            store[name]['x'].append(x)
            store[name]['y'].append(dsm.normalize_probabilities(y))
            # make prediction targets for individual song types [OPTIONAL]
            if make_single_class_datasets:
                for cnt, class_name in enumerate(class_names[1:]):
                    store[name][f'y_{class_name}'].append(dsm.normalize_probabilities(y[:, [0, cnt+1]]))
        else:
            # split data from each remaining file into train and test chunks according to `splits`
            if block_stratify:
                stratify = y
            else:
                stratify = None

            # convert block_size from seconds to samples
            if block_size is not None:
                block_size = int(block_size * fs)

            # FIXME: lenght of split does not match with fractions
            split_arrays = dsm.generate_data_splits({'x': x, 'y': y}, data_splits, data_split_targets, seed=seed,
                                                     block_stratify=stratify, block_size=block_size)
            logging.info(f'    splitting {data_splits} into {data_split_targets}.')
            for name in set(data_split_targets):
                store[name]['x'].append(split_arrays['x'][name])
                store[name]['y'].append(dsm.normalize_probabilities(split_arrays['y'][name]))
                # make prediction targets for individual song types [OPTIONAL]
                if make_single_class_datasets:
                    for cnt, class_name in enumerate(class_names[1:]):
                        store[name][f'y_{class_name}'].append(dsm.normalize_probabilities(split_arrays['y'][name][:, [0, cnt+1]]))

    # report
    logging.info(f"  Got {store['train']['x'].shape}, {store['val']['x'].shape}, {store['test']['x'].shape} train/val/test samples.")
    # save as npy_dir
    logging.info(f'  Saving to {store_folder}.')
    das.npy_dir.save(store_folder, store)
    if delete_intermediate_store:
        pass  # TODO delete intermediate store



"""Load FicTrac ball tracking data

should return:
    x: pd.DataFrame[frames, (variables)]
"""
import numpy as np
import pandas as pd
from .. import io
from typing import Optional
import logging
import xarray as xr

class BallTracks():

    def make(self, filename: Optional[str] = None):
        if filename is None:
            filename = self.path

        balltracks, first_balltracked_frame, last_balltracked_frame = self.load(filename)

        xr_balltracks = xr.DataArray(data=balltracks.values,
                                    dims=['frame_number_ball', 'data_ball'],
                                    coords={'frame_number_ball': balltracks['frame counter (starts at 1)']-1,
                                            'data_ball': list(balltracks.columns)},
                                    attrs={'loader': self.NAME,
                                           'kind': self.KIND,
                                           'path': filename,})
        return xr_balltracks


@io.register_provider
class FicTrac_balltracks(io.BaseProvider, BallTracks):

    KIND = 'balltracks'
    NAME = 'fictrac csv'
    SUFFIXES = ['_fictrac.csv']

    def load(self, filename: Optional[str] = None):
        """Load tracker data"""
        if filename is None:
            filename = self.path

        balltracks = pd.read_csv(filename, index_col=0)
        first_balltracked_frame = balltracks['frame counter (starts at 1)'].values[0]
        last_balltracked_frame = balltracks['frame counter (starts at 1)'].values[-1]
        return balltracks, first_balltracked_frame, last_balltracked_frame



import xarray_behave as xb
import logging

logging.getLogger().setLevel(logging.INFO)

recs = {'rpi9-20210409_093149': {},
        'localhost-20210617_113024': {},
        'localhost-20181120_144618': {},
        'localhost-20210624_104612': {},
        'localhost-20210628_145223': {'dat_path': 'dat', 'res_path': 'res', 'root': 'tests/data'},
        'localhost-20210629_171532': {'dat_path': 'dat', 'res_path': 'res', 'root': 'tests/data'},
        'Dmel_male': {'dat_path': 'dat', 'res_path': 'dat', 'root': 'tests/data', 'filepath_daq': 'dat/Dmel_male.wav', 'filepath_annotations': 'dat/Dmel_male_annotations.csv'},
        'Dmel_male2': {'dat_path': 'dat', 'res_path': 'dat', 'root': 'tests/data', 'filepath_daq': 'dat/Dmel_male.npz', 'filepath_annotations': 'dat/Dmel_male_annotations.csv'}}


# TODO: add asserts to ensure results conform to expected ds structure
def test_assemble2():
    datenames = list(recs.keys())
    ii = 2
    datename = datenames[ii]
    kwargs = recs[datename]
    ds = xb.assemble(datename, **kwargs)
    dm = xb.assemble_metrics(ds)


if __name__ == '__main__':
    test_assemble2()



from typing import Iterable
import cv2
import numpy as np
import h5py
import colorcet
from typing import Iterable

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtWidgets, QtCore
import logging
from videoreader import VideoReader
from typing import Union


def make_colors(nb_colors: int) -> Iterable:
    colors = []
    if nb_colors > 0:
        # cmap = colorcet.cm['glasbey_light']
        # cmap = colorcet.cm['glasbey_bw_minc_20_minl_50']
        cmap = colorcet.cm['glasbey_bw_minc_20_minl_30']
        # ignore first (red)
        colors = (cmap(np.arange(1, nb_colors+1)) * 255)[:, :3].astype(np.uint8)
    return colors


def fast_plot(plot_widget, x, y, pen=None):
    """[summary]

    Args:
        plot_widget ([type]): [description]
        x ([type]): [description]
        y ([type]): [description]
        pen ([type], optional): [description]. Defaults to None.
    """
    # as per: https://stackoverflow.com/questions/17103698/plotting-large-arrays-in-pyqtgraph
    # not worth it when plotting all channels but may speed up drawing events
    conn = np.ones_like(x, dtype=np.bool)
    conn[:, -1] = False  # make sure plots are disconnected
    path = pg.arrayToQPath(x.flatten(), y.flatten(), conn.flatten())
    item = QtGui.QGraphicsPathItem(path)
    if pen is None:
        pen = pg.mkPen(color=(196, 128, 128))
    item.setPen(pen)
    plot_widget.addItem(item)
    return item


class VideoReaderNP(VideoReader):
    """VideoReader posing as numpy array."""

    def __getitem__(self, index):
        return self.read(index)[1]

    @property
    def dtype(self):
        return np.uint8

    @property
    def shape(self):
        return (self.number_of_frames, *self.frame_shape)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def size(self):
        return np.product(self.shape)

    def min(self):
        return 0

    def max(self):
        return 255

    def transpose(self, *args):
        return self


class ImageViewVR(pg.ImageView):

    def quickMinMax(self, data):
        """Dummy min/max for numpy videoreader. The original function tries to read the full video!"""
        return 0, 255


class FastImageWidget(pg.GraphicsLayoutWidget):

    def __init__(self, *args, useOpenGL=True, **kwargs):
        """[summary]

        Args:
            useOpenGL (bool, optional): [description]. Defaults to True.
        """
        super(FastImageWidget, self).__init__(*args, **kwargs)
        self.viewBox = self.addViewBox()
        self.useOpenGL(useOpenGL)
        self.pixmapItem = QtWidgets.QGraphicsPixmapItem()
        self.pixmapItem.setShapeMode(QtWidgets.QGraphicsPixmapItem.BoundingRectShape)
        self.viewBox.addItem(self.pixmapItem)
        self.viewBox.setAspectLocked(lock=True, ratio=1)
        self.viewBox.disableAutoRange()

    def registerMouseClickEvent(self, func):
        """[summary]

        Args:
            func ([type]): [description]
        """
        self.pixmapItem.mouseClickEvent = func

    def setImage(self, image, image_format=QtGui.QImage.Format_RGB888, auto_scale: bool = False):
        """[summary]

        Args:
            image ([type]): [description]
            image_format ([type], optional): [description]. Defaults to QtGui.QImage.Format_RGB888.
            auto_scale (bool, optional): [description]. Defaults to False.
        """
        qimg = QtGui.QImage(image, image.shape[1], image.shape[0], image_format)
        # qimg = QtGui.QImage(image.ctypes.data, image.shape[1], image.shape[0], image_format)

        qpix = QtGui.QPixmap(qimg)
        self.pixmapItem.setPixmap(qpix)
        if auto_scale:
            self.fitImage(image.shape[1], image.shape[0])

    def fitImage(self, width: int, height: int):
        """[summary]

        Args:
            width ([type]): [description]
            height ([type]): [description]
        """
        self.viewBox.setRange(xRange=(0, width), yRange=(0, height), padding=0)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def allkeys(obj, keys=[]):
  """Recursively find all keys"""
  # from https://stackoverflow.com/questions/59897093/get-all-keys-and-its-hierarchy-in-h5-file-using-python-library-h5py
  keys.append(obj.name)
  if isinstance(obj, h5py.Group):
    for item in obj:
      if isinstance(obj[item], h5py.Group):
        allkeys(obj[item], keys)
      else: # isinstance(obj[item], h5py.Dataset):
        keys.append(obj[item].name)
  return keys


def is_sorted(array):
    return np.all(np.diff(array) >= 0)


def find_nearest_idx(array: np.array, values: Union[int, float, np.array]):
    """Find nearest index of each value from values in array

    from https://stackoverflow.com/a/46184652

    Args:
        array (np.array): array to search in
        values (Union[int, float, np.array]): query, should be sorted.

    Returns:
        [type]: indices of entries in array closest to each value in values
    """

    # scalar query
    if isinstance(values, float) or isinstance(values, int):
        return (np.abs(array - values)).argmin()

    # make sure array is a numpy array
    array = np.array(array)
    if not is_sorted(array):
        array = np.sort(array)

    # get insert positions
    idxs = np.searchsorted(array, values, side="left")

    # find indexes where previous index is closer
    prev_idx_is_less = ((idxs == len(array))|(np.fabs(values - array[np.maximum(idxs-1, 0)]) < np.fabs(values - array[np.minimum(idxs, len(array)-1)])))
    idxs[prev_idx_is_less] -= 1
    return idxs


class Worker(pg.QtCore.QRunnable):
    """Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    Args:
    fn: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    args: Arguments to pass to the callback function
    kwargs: Keywords to pass to the callback function
    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    @pg.QtCore.Slot()  # QtCore.Slot
    def run(self):
        '''
        Initialise the runner function with passed args, kwargs.
        '''
        self.fn(*self.args, **self.kwargs)


class InvokeEvent(QtCore.QEvent):
    EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self, fn, *args, **kwargs):
        QtCore.QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs


class Invoker(QtCore.QObject):
    def event(self, event):
        event.fn(*event.args, **event.kwargs)

        return True

_invoker = Invoker()


def invoke_in_main_thread(fn, *args, **kwargs):
    QtCore.QCoreApplication.postEvent(_invoker,
        InvokeEvent(fn, *args, **kwargs))


class CheckableComboBox(QtWidgets.QComboBox):
    # from https://gis.stackexchange.com/questions/350148/qcombobox-multiple-selection-pyqt5

    # Subclass Delegate to increase item height
    class Delegate(QtWidgets.QStyledItemDelegate):
        def sizeHint(self, option, index):
            size = super().sizeHint(option, index)
            size.setHeight(20)
            return size

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)

        # Use custom delegate
        self.setItemDelegate(CheckableComboBox.Delegate())

        # Update the text when an item is toggled
        self.model().dataChanged.connect(self.updateText)

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.closeOnLineEditClick = False

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

    def resizeEvent(self, event):
        # Recompute text to elide as needed
        self.updateText()
        super().resizeEvent(event)

    def eventFilter(self, object, event):

        if object == self.lineEdit():
            if event.type() == QtCore.QEvent.MouseButtonRelease:
                if self.closeOnLineEditClick:
                    self.hidePopup()
                else:
                    self.showPopup()
                return True
            return False

        if object == self.view().viewport():
            if event.type() == QtCore.QEvent.MouseButtonRelease:
                index = self.view().indexAt(event.pos())
                item = self.model().item(index.row())

                if item.checkState() == QtCore.Qt.Checked:
                    item.setCheckState(QtCore.Qt.Unchecked)
                else:
                    item.setCheckState(QtCore.Qt.Checked)
                return True
        return False

    def showPopup(self):
        super().showPopup()
        # When the popup is displayed, a click on the lineedit should close it
        self.closeOnLineEditClick = True

    def hidePopup(self):
        super().hidePopup()
        # Used to prevent immediate reopening when clicking on the lineEdit
        self.startTimer(100)
        # Refresh the display text when closing
        self.updateText()

    def timerEvent(self, event):
        # After timeout, kill timer, and reenable click on line edit
        self.killTimer(event.timerId())
        self.closeOnLineEditClick = False

    def updateText(self):
        texts = []
        for i in range(self.model().rowCount()):
            if self.model().item(i).checkState() == QtCore.Qt.Checked:
                texts.append(self.model().item(i).text())
        text = ", ".join(texts)

        # Compute elided text (with "...")
        metrics = QtGui.QFontMetrics(self.lineEdit().font())
        elidedText = metrics.elidedText(text, QtCore.Qt.ElideRight, self.lineEdit().width())
        self.lineEdit().setText(elidedText)

    def addItem(self, text, data=None):
        item = QtGui.QStandardItem()
        item.setText(text)
        if data is None:
            item.setData(text)
        else:
            item.setData(data)
        item.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable)
        item.setData(QtCore.Qt.Unchecked, QtCore.Qt.CheckStateRole)
        self.model().appendRow(item)

    def addItems(self, texts, datalist=None):
        for i, text in enumerate(texts):
            try:
                data = datalist[i]
            except (TypeError, IndexError):
                data = None
            self.addItem(text, data)

    def currentData(self):
        # Return the list of selected items data
        res = []
        for i in range(self.model().rowCount()):
            if self.model().item(i).checkState() == QtCore.Qt.Checked:
                res.append(self.model().item(i).data())
        return res



from typing import Dict, List, Union, Optional, Callable
import os
from glob import glob
from typing import Dict, List, Union, Optional, Callable
from collections import namedtuple, OrderedDict


class BaseProvider():

    KIND: str
    NAME: str
    SUFFIXES: List[str]

    def __init__(self, path: Optional[str] = None):
        self.path = path

    def info(self):
        return f"{self.NAME} loads {self.KIND} from files ending with {self.SUFFIXES}."

    @classmethod
    def match(cls, filename):
        if isinstance(filename, str):
            return any([filename.endswith(s) for s in cls.SUFFIXES])
        else:
            return False

    @classmethod
    def get_loader(cls, filename: str):
        if cls.match(filename):
            loader = cls(filename)
        else:
            loader = None
        return loader

    # def make(self, filename: Optional[str] = None, **kwargs):
    #     return None

    # def load(self, filename: Optional[str] = None, **kwargs):
    #     raise None


# TODO set precedence - maybe with provider.PRIORITY and sort providers accordingly?
providers = []
kinds = OrderedDict()
def register_provider(func):
    providers.append(func)
    if func.KIND not in kinds:
        kinds[func.KIND] = []
    kinds[func.KIND].extend(func.SUFFIXES)
    return func


def get_loader(kind: str, basename: str, stop_after_match: bool = True, basename_is_full_name: bool = False):
    """[summary]

    Args:
        kind (str): [description]
        basename (str): [description]
        stop_after_match (bool, optional): [description]. Defaults to True.
        basename_is_full_name (bool, optional): If False will append registered suffixes,
                                                otherwise will try to find match for basename as is.
                                                Defaults to False.

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    if kind not in kinds:
        raise ValueError(f'Unknow kind {kind} - must be in {kinds}.')

    loaders = []
    for suffix in kinds[kind]:
        if basename_is_full_name:  # use basename as is
            paths = glob(basename)
        else:  # add registered suffix
            paths = glob(basename + suffix)

        for path in paths:
            for provider in providers:
                loader = provider.get_loader(path)
                if loader is not None and loader.KIND == kind:
                    if stop_after_match:
                        return loader
                    else:
                        loaders.append(loader)
    return loaders


# need to import new modules here!
from . import (tracks,
               balltracks,
               annotations,
               annotations_manual,
               poses,
               audio,
               movieparams)


import pytest

import numpy as np
import xarray as xr
import pandas as pd


def test_imports():
    import xarray_behave as xb
    from xarray_behave import event_utils
    import xarray_behave.annot
    import xarray_behave.gui
    import xarray_behave.io



"""Load files created by various analysis programs created."""
import numpy as np
import pandas as pd
import h5py
from typing import Sequence, Optional, Tuple
import scipy.interpolate
import scipy.ndimage
import scipy.signal
import scipy.stats
from scipy.ndimage import maximum_filter1d
from samplestamps import SampStamp
import math


def rotate_point(point: Tuple[float, float], degrees: float, origin: Tuple[float, float] = (0, 0)) -> Tuple[float, float]:
    """Rotates 2D point around another point.

    Args:
        point (Tuple[float, float]): (x,y) coordinates of the point
        degrees (float): angle in degrees by which to rotate
        origin (Tuple[float, float], optional): point (x,y) around which to rotate point. Defaults to (0, 0).

    Returns:
        Tuple[float, float]: Rotated point
    """
    x, y = point
    radians = degrees / 180 * np.pi
    offset_x, offset_y = origin
    adjusted_x = (x - offset_x)
    adjusted_y = (y - offset_y)
    cos_rad = math.cos(radians)
    sin_rad = math.sin(radians)
    qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
    qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y
    return qx, qy


def rotate_pose(positions, degree, origin=(0, 0)):
    """Rotates point set.

    Args:
        pos (np.ndarray): [points, x/y]
        degrees (float): degree by which to rotate
        origin (Tuple[float, float], optional): point (x,y) around which to rotate point. Defaults to (0, 0).

    Returns:
        tuple: (x, y) rotated
    """
    radians = degree / 180 * np.pi
    cos_rad = np.cos(radians)
    sin_rad = np.sin(radians)
    rot = np.array([[cos_rad, sin_rad], [-sin_rad, cos_rad]])
    positions_centered = positions - origin
    return rot.dot(positions_centered.T).T + origin


def find_nearest(array, values):
    """Find nearest occurrence of each item of values in array.

    Args:
        array: find nearest in this list
        values: queries

    Returns:
        val: nearest val in array to each item in values
        idx: index of nearest val in array to each item in values
        dist: distance to nearest val in array for each item in values
        NOTE: Returns nan-arrays of the same size as values if `array` is empty.
    """
    if len(values) and len(array):  # only do this if boh inputs are non-empty lists
        values = np.atleast_1d(values)
        abs_dist = np.abs(np.subtract.outer(array, values))  # this eats a lot of memory - maybe use nearest neighbour interpolation for upsampling?
        idx = abs_dist.argmin(0)
        dist = abs_dist.min(0)
        val = array[idx]
    else:
        idx = np.full_like(values, fill_value=np.nan)
        dist = np.full_like(values, fill_value=np.nan)
        val = np.full_like(values, fill_value=np.nan)
    return val, idx, dist


def merge_channels(data, sampling_rate, filter_data: bool = True):
    """Merge channels based on a running maximum.

    Args:
        data (ndarray): [samples, channels]
        sampling_rate (num): in Hz

    Returns:
        ndarray: merged across
    """
    data = np.array(data)  # ensure data is an np.array (and not dask) - otherwise np.interp will fail
    mask = ~np.isfinite(data)  # remove all nan/inf data
    data[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), data[~mask])
    # band-pass filter out noise on each channel
    b, a = scipy.signal.butter(6, (25, 1500), btype='bandpass', fs=sampling_rate)
    data = scipy.signal.filtfilt(b, a, data, axis=0, method='pad')
    # find loudest channel in 101-sample windows
    if filter_data:
        sng_max = maximum_filter1d(np.abs(data), size=101, axis=0)
        loudest_channel = np.argmax(sng_max, axis=-1)
    # get linear index and merge channels
    idx = np.ravel_multi_index((np.arange(sng_max.shape[0]), loudest_channel), data.shape)
    data_merged_max = data.ravel()[idx]
    data_merged_max = data_merged_max[:, np.newaxis]  # shape needs to be [nb_samples, 1]
    return data_merged_max


def fill_gaps(sine_pred, gap_dur=100):
    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets<offsets[-1]]
        offsets = offsets[offsets>onsets[0]]
        durations = offsets - onsets
        for idx, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
            if idx>0 and offsets[idx-1]>onsets[idx]-gap_dur:
                sine_pred[offsets[idx-1]:onsets[idx]+1] = 1
    return sine_pred


def remove_short(sine_pred, min_len=100):
    # remove too short sine songs
    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets<offsets[-1]]
        offsets = offsets[offsets>onsets[0]]
        durations = offsets - onsets
        for cnt, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
            if duration<min_len:
                sine_pred[onset:offset+1] = 0
    return sine_pred


def load_swap_indices(filepath):
    a = np.loadtxt(filepath)  #, dtype=np.uintp)
    indices, flies1, flies2 = np.split(a, [1, 2], axis=-1)  # split columns in file into variables
    if indices.ndim == 2:  # is sometimes 2d for some reason
        indices = indices[:, 0]
    return indices, flies1.astype(np.uintp), flies2.astype(np.uintp)


def swap_flies(dataset, swap_times, flies1=0, flies2=1):
    """Swap flies in dataset.

    Caution: datavariables are currently hard-coded!
    Caution: Swap *may* be in place so *might* will alter original dataset.

    Args:
        dataset ([type]): Dataset for which to swap flies
        swap_times ([type]): List of times (in seconds) at which to swap flies.
        flies1 (int or list/tuple, optional): Either a single value for all indices or a list with one value per item in indices. Defaults to 0.
        flies2 (int or list/tuple, optional): Either a single value for all indices or a list with one value per item in indices. Defaults to 1.

    Returns:
        dataset with swapped indices ()
    """
    if 'time' not in dataset:
        return

    for cnt, swap_time in enumerate(swap_times):
        if isinstance(flies1, (list, tuple)) and isinstance(flies2, (list, tuple)):
            fly1, fly2 = flies1[cnt], flies2[cnt]
        else:
            fly1, fly2 = flies1, flies2

        nearest_time = dataset.time.sel(time=swap_time, method="nearest")

        index = int(np.where(dataset.time == nearest_time)[0])
        if 'pose_positions_allo' in dataset:
            dataset.pose_positions_allo.values[index:, [fly2, fly1], ...] = dataset.pose_positions_allo.values[index:, [fly1, fly2], ...]
        if 'pose_positions' in dataset:
            dataset.pose_positions.values[index:, [fly2, fly1], ...] = dataset.pose_positions.values[index:, [fly1, fly2], ...]
        if 'body_positions' in dataset:
            dataset.body_positions.values[index:, [fly2, fly1], ...] = dataset.body_positions.values[index:, [fly1, fly2], ...]

    return dataset


def load_raw_song(filepath_daq: str, song_channels: Optional[Sequence[int]] = None,
                  return_nonsong_channels: bool = False, lazy: bool = False):
    """[summary]

    Args:
        filepath_daq ([type]): [description]
        song_channels (List[int], optional): Sequence of integers as indices into 'samples' datasaet.
                                             Defaults to [0,..., 15].
        return_nonsong_channels (bool, optional): will return the data not in song_channels as separate array. Defaults to False
        lazy (bool, optional): If True, will load song as dask.array, which allows lazy indexing.
                               Otherwise, will load the full recording from disk (slow). Defaults to False.

    Returns:
        [type]: [description]
    """
    if song_channels is None:  # the first 16 channels in the data are the mic recordings
        song_channels = np.arange(16)

    if lazy:
        f = h5py.File(filepath_daq, mode='r', rdcc_w0=0, rdcc_nbytes=100 * (1024 ** 2), rdcc_nslots=50000)
        # convert to dask array since this allows lazily evaluated indexing...
        import dask.array as daskarray
        da = daskarray.from_array(f['samples'], chunks=(10000, 1))
        nb_channels = f['samples'].shape[1]
        song_channels = song_channels[song_channels < nb_channels]
        song = da[:, song_channels]
        if return_nonsong_channels:
            non_song_channels = list(set(list(range(nb_channels))) - set(song_channels))
            non_song = da[:, non_song_channels]
    else:
        with h5py.File(filepath_daq, 'r') as f:
            nb_channels = f['samples'].shape[1]
            song = f['samples'][:, song_channels]
            if return_nonsong_channels:
                non_song_channels = list(set(list(range(nb_channels))) - set(song_channels))
                non_song = f['samples'][:, non_song_channels]

    if return_nonsong_channels:
        return song, non_song
    else:
        return song


def load_times(filepath_timestamps, filepath_daq):
    """Load daq and cam time stamps, create muxer"""
    with h5py.File(filepath_timestamps, 'r') as f:
        cam_stamps = f['timeStamps'][:]

    # time stamps at idx 0 can be a little wonky - so use the information embedded in the image
    if cam_stamps.shape[1] > 2:  # time stamps from flycapture (old point grey API)
        shutter_times = cam_stamps[:, 1] + cam_stamps[:, 2]/1_000_000  # time of "Shutter OFF"
    elif cam_stamps.shape[1] == 2:  # time stamps from other camera drivers (spinnaker (new point grey/FLIR API) and ximea)
        # cut off empty time stamps
        last_frame_idx = np.argmax(cam_stamps[:, 1]==0) - 1
        cam_stamps = cam_stamps[:last_frame_idx]

        # fix jumps from overflow in timestamp counter for ximea cameras
        frame_intervals = np.diff(cam_stamps[:, 1])
        frame_interval_median = np.median(frame_intervals)
        idxs = np.where(frame_intervals<-10 * frame_interval_median)[0]
        while len(idxs):
            idx = idxs[0]
            df_wrong = cam_stamps[idx+1, 1] - cam_stamps[idx, 1]
            df_inferred = cam_stamps[idx+1, 0] - cam_stamps[idx, 0]
            cam_stamps[idx+1:, 1] = cam_stamps[idx+1:, 1] - df_wrong + df_inferred

            frame_intervals = np.diff(cam_stamps[:, 1])
            idxs = np.where(frame_intervals<-10 * frame_interval_median)[0]

        shutter_times = cam_stamps[:, 1]

    last_frame_idx = np.argmax(shutter_times==0) - 1
    shutter_times = shutter_times[:last_frame_idx]

    # DAQ time stamps
    with h5py.File(filepath_daq, 'r') as f:
        daq_stamps = f['systemtime'][:]
        daq_sampleinterval = f['samplenumber'][:]

    # remove trailing zeros - may be left over if recording didn't finish properly
    if 0 in daq_stamps:
        last_valid_idx = np.argmax(daq_stamps == 0)
    else:
        last_valid_idx = len(daq_stamps) - 1  # in case there are no trailing zeros
    daq_samplenumber = np.cumsum(daq_sampleinterval)[:last_valid_idx, np.newaxis]
    last_sample = daq_samplenumber[-1, 0]

    nb_seconds_per_interval, _ = scipy.stats.mode(np.diff(daq_stamps[:last_valid_idx, 0]))  # seconds - using mode here to be more robust
    nb_seconds_per_interval = nb_seconds_per_interval[0]
    nb_samples_per_interval = np.mean(np.diff(daq_samplenumber[:last_valid_idx, 0]))
    sampling_rate_Hz = np.around(nb_samples_per_interval / nb_seconds_per_interval, -3)  # round to 1000s of Hz

    ss = SampStamp(sample_times=daq_stamps[:last_valid_idx, 0], frame_times=shutter_times, sample_numbers=daq_samplenumber[:, 0], auto_monotonize=False)
    # # different refs:
    #
    # # first sample is 0 seconds
    # s0 = ss.sample_time(0)
    # ss = SampStamp(sample_times=daq_stamps[:, 0] - s0, frame_times=cam_stamps[:, 0] - s0, sample_numbers=daq_samplenumber[:, 0])
    #
    # # first frame is 0 seconds - for no-resample-video-data
    # f0 = ss.frame_time(0)
    # ss = SampStamp(sample_times=daq_stamps[:, 0] - f0, frame_times=cam_stamps[:, 0] - f0, sample_numbers=daq_samplenumber[:, 0])

    return ss, last_sample, sampling_rate_Hz


def load_movietimes(filepath_timestamps, filepath_daq):
    """Load daq and cam time stamps, create muxer"""
    df = pd.read_csv(filepath_timestamps)

    # DAQ time stamps
    with h5py.File(filepath_daq, 'r') as f:
        daq_stamps = f['systemtime'][:]
        daq_sampleinterval = f['samplenumber'][:]

    # remove trailing zeros - may be left over if recording didn't finish properly
    if 0 in daq_stamps:
        last_valid_idx = np.argmax(daq_stamps == 0)
    else:
        last_valid_idx = len(daq_stamps) - 1  # in case there are no trailing zeros
    daq_samplenumber = np.cumsum(daq_sampleinterval)[:last_valid_idx, np.newaxis]
    last_sample = daq_samplenumber[-1, 0]

    nb_seconds_per_interval, _ = scipy.stats.mode(np.diff(daq_stamps[:last_valid_idx, 0]))  # seconds - using mode here to be more robust
    nb_seconds_per_interval = nb_seconds_per_interval[0]
    nb_samples_per_interval = np.mean(np.diff(daq_samplenumber[:last_valid_idx, 0]))
    sampling_rate_Hz = np.around(nb_samples_per_interval / nb_seconds_per_interval, -3)  # round to 1000s of Hz


    # ss = SampStamp(sample_times=daq_stamps[:last_valid_idx, 0], frame_times=shutter_times, sample_numbers=daq_samplenumber[:, 0], auto_monotonize=False)
    ss = SampStamp(sample_times=daq_stamps[:last_valid_idx, 0], sample_numbers=daq_samplenumber[:, 0],
                   frame_samples=df['sample'], frame_numbers=df['movie_frame'], auto_monotonize=False)
    # # different refs:
    #
    # # first sample is 0 seconds
    # s0 = ss.sample_time(0)
    # ss = SampStamp(sample_times=daq_stamps[:, 0] - s0, frame_times=cam_stamps[:, 0] - s0, sample_numbers=daq_samplenumber[:, 0])
    #
    # # first frame is 0 seconds - for no-resample-video-data
    # f0 = ss.frame_time(0)
    # ss = SampStamp(sample_times=daq_stamps[:, 0] - f0, frame_times=cam_stamps[:, 0] - f0, sample_numbers=daq_samplenumber[:, 0])

    return ss, last_sample, sampling_rate_Hz


def fix_keys(d):
    d_new = dict()
    # HACK zarr (or xarray) cuts off long string keys in event-types
    fix_dict = {'aggression_manu': 'aggression_manual', 'vibration_manua': 'vibration_manual'}
    # make this a function!!
    for eventtype in d.keys():
        # convert all from b'..' to str
        try:
            d_new[eventtype.decode()] = d[eventtype]
        except AttributeError:
            d_new[eventtype] = d[eventtype]

    for eventtype in d_new:
        if eventtype in fix_dict.keys():
            # logging.info(f'   Renaming {eventtype} to {fix_dict[eventtype]}.')
            d_new[fix_dict[eventtype]] = d_new.pop(eventtype)
    return d_new



"""
Widgets and dialogues for YAML-based forms.

Most of the time you can use :py:class:`YamlFormWidget`. If you want to show
the widget as a model dialog, :py:class:`FormBuilderModalDialog` makes this
a little more convenient (it provides methods for adding a message for the
dialog and for getting the results when the dialog is closed).

Example of form widget:

>>> widget = YamlFormWidget(yaml_file="example.yaml")
>>> widget.mainAction.connect(my_function)

my_function will get called with form data when user clicks the main button
(main button has type "button" and default "main action")

Example of modal dialog:

>>> results = FormBuilderModalDialog(form_name="example").get_results()

The results will be empty dictionary if the user hit "cancel", otherwise it
will contain all data from form (dict keys matching names of fields).

The logic which creates each form field based on the data from the YAML file
is :py:method:`FormBuilderLayout.add_item()`. Look there if you want to know
what's supported in the YAML file, what exactly each field type does, or if you
want to add a new type of supported form field.
"""



# from PySide2 import QtWidgets, QtCore
# modified from https://sleap.ai/_modules/sleap/gui/formbuilder.html
import yaml
from typing import Any, Dict, List, Optional, Union, Text
try:
    import PySide2  # this will force pyqtgraph to use PySide instead of PyQt4/5
except ImportError:
    pass
from pyqtgraph.Qt import QtWidgets, QtCore, QtGui


class YamlDialog(QtGui.QDialog):
    def __init__(self, yaml_file, parent=None, title=None,
                 main_callback=None, callbacks={}):
        super().__init__(parent=parent)
        self.setWindowTitle(title)
        self.form = YamlFormWidget(yaml_file)
        if main_callback is None:
            def main_callback():
                """Accepts and closes dialog."""
                self.accept()
                self.close()
        self.form.mainAction.connect(main_callback)

        self.scroll_area = QtGui.QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        # self.scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        # self.scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidget(self.form)


        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.scroll_area)
        for button, callback in callbacks.items():
            self.form.buttons[button].clicked.connect(callback)
        self.setLayout(layout)

        self.resize(650, 800)


class YamlFormWidget(QtWidgets.QGroupBox):
    """
    Widget which shows form created from a YAML file.

    Typically you'll want to save the YAML in `sleap/config/` and use the
    :py:meth:`from_name` method to make the form (e.g., if your form data is in
    `sleap/config/foo.yaml`, then you can create form like so:

    >>> widget = YamlFormWidget.from_name("foo")

    Args:
        yaml_file: filename of YAML file to load, or dict.
        which_form (optional): key to form in YAML file, default is "main".
            this allows a single YAML file to contain data for multiple forms.
    """

    mainAction = QtCore.Signal(dict)
    valueChanged = QtCore.Signal()

    def __init__(
        self,
        yaml_file: Union[Text, Dict],
        which_form: Text = "main",
        field_options_lists: Optional[Dict[Text, list]] = None,
        *args,
        **kwargs,
    ):
        super(YamlFormWidget, self).__init__(*args, **kwargs)
        if isinstance(yaml_file, dict):
            items_to_create = yaml_file
        else:
            with open(yaml_file, "r") as form_yaml:
                items_to_create = yaml.load(form_yaml, Loader=yaml.SafeLoader)

        self.which_form = which_form

        self.form_layout = FormBuilderLayout(
            items_to_create[self.which_form], field_options_lists=field_options_lists,
        )

        self.setLayout(self.form_layout)

        if items_to_create[self.which_form]:
            for item in items_to_create[self.which_form]:
                if (
                    item["type"] == "button"
                    and item.get("default", "") == "main action"
                ):
                    self.buttons[item["name"]].clicked.connect(self.trigger_main_action)

        self.form_layout.valueChanged.connect(self.valueChanged)

    def __getitem__(self, key):
        """Return value for specified form field."""
        return FormBuilderLayout.get_widget_value(self.fields[key])

    def __setitem__(self, key, val):
        """Set value for specified form field."""
        FormBuilderLayout.set_widget_value(self.fields[key], val)

    @property
    def buttons(self):
        """Returns a list of buttons in form (so we can connect to handlers)."""
        return self.form_layout.buttons

    @property
    def fields(self):
        """Return a dict of {name: widget} fields in form."""
        return self.form_layout.fields

    def set_form_data(self, data):
        """Set data for form from dict."""
        self.form_layout.set_form_data(data)

    def get_form_data(self):
        """Returns dict of form data."""
        return self.form_layout.get_form_data()

    def set_field_options(self, field_name: Text, options_list: List[Text], **kwargs):
        """Sets option list for specified field."""
        self.form_layout.set_field_options(field_name, options_list, **kwargs)

    def set_field_enabled(self, field_name: Text, is_enabled: bool):
        self.form_layout.set_field_enabled(field_name, is_enabled)

    def set_enabled(self, enabled: bool):
        self.form_layout.setEnabled(enabled)

    def trigger_main_action(self):
        """Emit mainAction signal with form data."""
        self.mainAction.emit(self.get_form_data())


class FormBuilderModalDialog(QtWidgets.QDialog):
    """
    Blocking modal dialog to use with :py:class:`YamlFormWidget` widget.

    You can either initialize with a :py:class:`YamlFormWidget` widget, or
    provide the name of the YAML form (i.e., the string you'd pass to
    :py:meth:`YamlFormWidget.from_name()`).
    """

    def __init__(
        self,
        form_widget: Optional[YamlFormWidget] = None,
        form_name: Optional[Text] = None,
        *args,
        **kwargs,
    ):
        super(FormBuilderModalDialog, self).__init__()

        if not form_widget and form_name:
            form_widget = YamlFormWidget.from_name(form_name)

        if not form_widget:
            raise ValueError(
                "FormBuilderModalDialog must have either form widget or name."
            )

        self._results = None
        self.form_widget = form_widget
        self.message_fields = []

        # Layout for buttons
        buttons = QtWidgets.QDialogButtonBox()
        self.cancel_button = buttons.addButton(QtWidgets.QDialogButtonBox.Cancel)
        self.run_button = buttons.addButton(QtWidgets.QDialogButtonBox.Ok)

        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addWidget(buttons, alignment=QtCore.Qt.AlignTop)

        buttons_layout_widget = QtWidgets.QWidget()
        buttons_layout_widget.setLayout(buttons_layout)

        # Layout for entire dialog
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.form_widget)
        layout.addWidget(buttons_layout_widget)

        self.setLayout(layout)

        # Connect actions for buttons
        buttons.accepted.connect(self.on_accept)
        buttons.rejected.connect(self.reject)

    def add_message(self, message: Text):
        """Adds text message between form fields and buttons."""
        field = QtWidgets.QLabel(message)
        field.setWordWrap(True)

        self.message_fields.append(field)

        self.layout().insertWidget(1, field)

    def set_message(self, message: Text):
        """Adds/replaces text message between form fields and buttons."""
        if self.message_fields:
            self.message_fields[0].setText(message)
        else:
            self.add_message(message)

    def on_accept(self):
        self._results = self.form_widget.get_form_data()
        self.accept()

    def get_results(self) -> Optional[Dict[Text, Any]]:
        """Shows dialog, blocks till submitted, returns dict of form data."""
        self._results = None
        self.exec_()
        return self._results


class FormBuilderLayout(QtWidgets.QFormLayout):
    """
    Custom `QFormLayout` which populates itself from list of form fields.

    Args:
        items_to_create: list of form items, each item in list is a dictionary
            with information about item. see :py::meth:`build_form` for details
            about the keys/values. typically this list comes straight from a
            YAML file.
        field_options_lists: allows you to set list of options to use for "list"
            fields (i.e., :py:class:`FieldComboWidget` or
            :py:class:`TextOrListWidget` widgets)
            when creating the form programmatically, instead of having options
            as "options" key of dictionary (which typically would be hard-coded
            in YAML).
    """

    valueChanged = QtCore.Signal()

    def __init__(
        self,
        items_to_create: List[Dict[Text, Any]],
        field_options_lists: Optional[Dict[Text, List[Text]]] = None,
        *args,
        **kwargs,
    ):
        super(FormBuilderLayout, self).__init__(*args, **kwargs)

        self.form_text_widget = None

        self.buttons = dict()
        self.fields = dict()
        self.field_options_lists = field_options_lists or dict()
        self.build_form(items_to_create)

    @property
    def stacked(self) -> list:
        """List of all "stack" widgets in form."""
        return [w for w in self.fields.values() if type(w) == StackBuilderWidget]

    def setEnabled(self, enabled: bool):
        for field in self.fields.values():
            if hasattr(field, "set_fields_enabled"):
                field.set_fields_enabled(enabled)
            if hasattr(field, "setEnabled"):
                field.setEnabled(enabled)

    def get_form_data(self) -> dict:
        """Gets all user-editable data from the widgets in the form layout.

        Returns:
            Dict with key:value for each user-editable widget in layout
        """
        widgets = self.fields.values()
        data = {
            w.objectName(): self.get_widget_value(w)
            for w in widgets
            if len(w.objectName())
            and type(w) not in (QtWidgets.QLabel, QtWidgets.QPushButton)
        }

        stacked_data = [w.get_data() for w in widgets if type(w) == StackBuilderWidget]
        for stack in stacked_data:
            data.update(stack)
        return data

    def set_form_data(self, data: Dict[Text, Any]):
        """Set specified user-editable data.

        Args:
            data: dictionary of data, key should match field name
        """
        widgets = self.fields
        for name, val in data.items():
            # print(f"Attempting to set {name} to {val}")
            if name in widgets:
                # print(f"set {name} to {val} ({type(val)})")
                self.set_widget_value(widgets[name], val)
            else:
                # print(f"cannot set {name} for {widgets}")
                pass

        for stacked_widget in self.stacked:
            stacked_widget.set_form_data(data)

    def set_field_options(self, field_name: Text, options_list: List[Text]):
        """Sets custom list of options for specified field."""
        self.field_options_lists[field_name] = options_list

        for subform in self.stacked:
            subform.set_field_options(field_name, options_list)

        self.update_field_options()

    def update_field_options(self):
        """Updates options list for every field with custom list."""
        for field_name, field in self.fields.items():
            if field_name in self.field_options_lists:
                field.set_options(self.field_options_lists[field_name])

    def set_field_enabled(self, field_name: Text, is_enabled: bool):
        """Sets whether the field is enabled/disabled."""
        fields = self.find_field(field_name)

        for field in fields:
            field.setEnabled(is_enabled)

    def find_field(self, field_name: Text):
        """
        Finds form fields by name.

        Args:
            field_name: Name of field to find.

        Returns:
            List of field widgets, including any in active stacked widget.
        """
        widgets = self.fields.values()
        found = [
            w
            for w in widgets
            if w.objectName() == field_name
            and type(w) not in (QtWidgets.QLabel, QtWidgets.QPushButton)
        ]
        stacks = [w for w in widgets if type(w) == StackBuilderWidget]
        for stack in stacks:
            found.extend(stack.find_field(field_name))
        return found

    @staticmethod
    def set_widget_value(widget: QtWidgets.QWidget, val):
        """Set value for specific widget."""

        if hasattr(widget, "isChecked"):
            widget.setChecked(val)
        elif hasattr(widget, "setValue"):
            widget.setValue(val)
        elif hasattr(widget, "currentText"):
            widget.setCurrentText(str(val))
        elif hasattr(widget, "text"):
            widget.setText(str(val))
        else:
            print(f"don't know how to set value for {widget}")
        # for macOS we need to call repaint (bug in Qt?)
        widget.repaint()

    @staticmethod
    def get_widget_value(widget: QtWidgets.QWidget) -> Any:
        """Returns value of form field.

        This determines the method appropriate for the type of widget.

        Args:
            widget: The widget for which to return value.
        Returns:
            value (can be bool, numeric, string, or None)
        """
        if hasattr(widget, "isChecked"):
            val = widget.isChecked()
        elif hasattr(widget, "value"):
            val = widget.value()
        elif hasattr(widget, "currentText"):
            val = widget.currentText()
        elif hasattr(widget, "text"):
            val = widget.text()
        elif hasattr(widget, "currentIndex"):
            val = widget.currentIndex()
        else:
            val = None
        if widget.property("field_data_type") == "sci":
            val = float(val)
        elif widget.property("field_data_type") == "int":
            val = int(val)
        elif widget.property("field_data_type").startswith("file_"):
            val = None if val == "None" else val
        return val

    def build_form(self, items_to_create: List[Dict[Text, Any]]):
        """Adds widgets to form layout for each item in items_to_create.

        Args:
            items_to_create: list of dictionaries with keys

              * name: used as key when we return form data as dict
              * label: string to show in form
              * type: supports double, int, bool, list, button, stack
              * default: default value for form field
              * [options]: comma separated list of options,
                used for list or stack field-types
              * for stack, array of dicts w/ form data for each stack page

        A "stack" has a dropdown menu that determines which stack page to show.

        Returns:
            None.
        """
        if not items_to_create:
            return
        for item in items_to_create:
            self.add_item(item)

    def add_item(self, item: Dict[Text, Any]):
        if item["type"] == "text":
            field = QtWidgets.QLabel(item["text"])
            field.setWordWrap(True)

            # We don't need to keep track of this text-only field so we'll
            # add it to the form and skip the other things we usually do.
            self.addRow(field)
            return

        # double: show spinbox (number w/ up/down controls)
        elif item["type"] == "double":
            field = QtWidgets.QDoubleSpinBox()

            min, max = -1000, 1000
            if "range" in item.keys():
                min, max = list(map(float, item["range"].split(",")))
            field.setRange(min, max)
            field.setSingleStep(0.25)

            decimals = item.get('decimals', 2)
            field.setDecimals(decimals)

            field.setValue(item["default"])

            field.valueChanged.connect(lambda: self.valueChanged.emit())

        # int: show spinbox (number w/ up/down controls)
        elif item["type"] == "int":
            field = QtWidgets.QSpinBox()
            if "range" in item.keys():
                min, max = list(map(int, item["range"].split(",")))
                field.setRange(min, max)
            elif item["default"] > 100:
                min, max = 0, item["default"] * 10
                field.setRange(min, max)
            field.setValue(item["default"])
            field.valueChanged.connect(lambda: self.valueChanged.emit())

        elif item["type"] in ("optional_int", "optional_double", "auto_int"):
            spin_type = item["type"].split("_")[-1]
            none_string = "auto" if item["type"].startswith("auto") else "none"
            none_label = item.get("none_label", None)
            decimals = item.get('decimals', 2)

            field = OptionalSpinWidget(
                type=spin_type, none_string=none_string, none_label=none_label, decimals=decimals
            )
            if "range" in item.keys():
                caster = int if spin_type=="int" else float
                min, max = list(map(caster, item["range"].split(",")))
                field.setRange(min, max)
            elif item["default"] is not None and item["default"] > 100:
                min, max = 0, item["default"] * 10
                field.setRange(min, max)

            field.setValue(item["default"])
            if item.get("default_disabled", False):
                field.setToNone()

            field.valueChanged.connect(lambda: self.valueChanged.emit())

        # bool: show checkbox
        elif item["type"] == "bool":
            field = QtWidgets.QCheckBox()
            field.setChecked(item["default"])
            field.stateChanged.connect(lambda: self.valueChanged.emit())

        # list: show drop-down menu
        elif item["type"] in ("list", "optional_list"):
            type_options = item.get("type-options", "")

            result_as_optional_idx = False
            add_blank_option = False
            if type_options == "optional_index":
                result_as_optional_idx = True
                add_blank_option = True
            if item["type"] == "optional_list":
                add_blank_option = True

            field = TextOrListWidget(
                result_as_idx=result_as_optional_idx, add_blank_option=add_blank_option,
            )

            if item["name"] in self.field_options_lists:
                field.set_options(self.field_options_lists[item["name"]])
            elif "options" in item:
                field.set_options(
                    item["options"].split(","), select_item=item.get("default", "")
                )

            field.valueChanged.connect(lambda: self.valueChanged.emit())

        # button
        elif item["type"] == "button":
            field = QtWidgets.QPushButton(item["label"])
            self.buttons[item["name"]] = field

        # string
        elif item["type"] in ("string", "optional_string"):
            field = QtWidgets.QLineEdit()
            val = item.get("default", "")
            val = "" if val is None else val
            field.setText(str(val))

        elif item["type"] == "string_list":
            field = StringListWidget()
            val = item.get("default", "")
            field.setValue(val)

        # stacked: show menu and form panel corresponding to menu selection
        elif item["type"] == "stacked":
            field = StackBuilderWidget(
                item, field_options_lists=self.field_options_lists
            )
            field.valueChanged.connect(lambda: self.valueChanged.emit())

        # If we don't recognize the type, just show a text box
        else:
            field = QtWidgets.QLineEdit()
            field.setText(str(item.get("default", "")))
            if item["type"].split("_")[0] == "file":
                field.setDisabled(True)

        # Store name and type on widget
        field.setObjectName(item["name"])
        field.setProperty("field_data_type", item.get("dtype", item["type"]))

        # Set tooltip for field
        if "help" in item:
            field.setToolTip(item["help"])

        # Store widget by name
        self.fields[item["name"]] = field

        # Add field (and label if appropriate) to form layout
        if item["type"] in ("stacked"):
            self.addRow(field)
        elif item["type"] in ("button"):
            self.addRow("", field)
        else:
            self.addRow(item["label"] + ":", field)

        # file_[open|dir]: show button to select file/directory
        if item["type"].split("_")[0] == "file":
            self.addRow("", self._make_file_button(item, field))

    def _make_file_button(
        self, item: Dict, field: QtWidgets.QWidget
    ) -> QtWidgets.QPushButton:
        """Creates the button for a file_* field-type."""
        file_button = QtWidgets.QPushButton("Select " + item["label"])

        if item["type"].split("_")[-1] == "open":
            # Define function for button to trigger
            def select_file(*args, x=field):
                filter = item.get("filter", "Any File (*.*)")
                filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                    None, caption="Open File", filter=filter
                )
                if len(filename):
                    x.setText(filename)
                self.valueChanged.emit()

        elif item["type"].split("_")[-1] == "dir":
            # Define function for button to trigger
            def select_file(*args, x=field):
                filename = QtWidgets.QFileDialog.getExistingDirectory(None, caption="Open File")
                if len(filename):
                    x.setText(filename)
                self.valueChanged.emit()

        else:
            select_file = lambda: print(f"no action set for type {item['type']}")

        file_button.clicked.connect(select_file)
        return file_button


class StackBuilderWidget(QtWidgets.QWidget):
    """
    Widget that shows different subforms depending on menu selection.

    Arguments:
        stack_data: Dictionary for field from `items_to_create`.
            The "options" key will give the list of options to show in
            menu. Each of the "options" will also be the key of a dictionary
            within stack_data that has the same structure as the dictionary
            passed to :py:meth:`FormBuilderLayout.build_form()`.
        field_options_list: passed to stacked forms, see
            :py:class:`FormBuilderLayout` for details.
    """

    valueChanged = QtCore.Signal()

    def __init__(self, stack_data, field_options_lists=None, *args, **kwargs):
        super(StackBuilderWidget, self).__init__(*args, **kwargs)

        self.option_list = stack_data["options"].split(",")

        self.field_options_lists = field_options_lists or None

        multi_layout = QtWidgets.QFormLayout()
        multi_layout.setFormAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        self.combo_box = QtWidgets.QComboBox()
        self.stacked_widget = ResizingStackedWidget()

        self.combo_box.activated.connect(self.switch_to_idx)

        self.page_layouts = dict()

        for page in self.option_list:
            # add page
            self.page_layouts[page] = FormBuilderLayout(
                stack_data[page], field_options_lists=self.field_options_lists
            )

            self.page_layouts[page].valueChanged.connect(self.valueChanged)

            page_widget = QtWidgets.QGroupBox()
            page_widget.setLayout(self.page_layouts[page])

            self.stacked_widget.addWidget(page_widget)

            # add option to menu
            self.combo_box.addItem(page)

        if len(stack_data.get("label", "")):
            combo_label = f"{stack_data['label']}:"
        else:
            combo_label = ""

        multi_layout.addRow(combo_label, self.combo_box)
        multi_layout.addRow(self.stacked_widget)

        self.setLayout(multi_layout)

        self.setValue(stack_data["default"])

    def switch_to_idx(self, idx):
        """Switch currently shown widget from stack."""
        self.stacked_widget.setCurrentIndex(idx)
        # Only show if the widget contains more than an empty layout
        if len(self.stacked_widget.currentWidget().children()) > 1:
            self.stacked_widget.show()
        else:
            self.stacked_widget.hide()
        self.valueChanged.emit()

    def value(self):
        """Returns value of menu."""
        return self.combo_box.currentText()

    def setValue(self, value):
        """Sets value of menu."""
        if value not in self.option_list:
            return
        idx = self.option_list.index(value)
        self.combo_box.setCurrentIndex(idx)
        self.switch_to_idx(idx)

    def setEnabled(self, val):
        self.combo_box.setEnabled(val)

    def set_fields_enabled(self, val):
        for subform in self.page_layouts.values():
            subform.setEnabled(val)

    def get_data(self):
        """Returns value from currently shown subform."""
        return self.page_layouts[self.value()].get_form_data()

    def find_field(self, *args, **kwargs):
        """Returns result of find_field method on currently shown subform."""
        return self.page_layouts[self.value()].find_field(*args, **kwargs)

    def set_field_options(self, *args, **kwargs):
        """Calls set_field_options for every subform."""
        for subform in self.page_layouts.values():
            subform.set_field_options(*args, **kwargs)

    def set_form_data(self, data):
        for layout in self.page_layouts.values():
            layout.set_form_data(data)


class OptionalSpinWidget(QtWidgets.QWidget):
    """
    Numeric (spin) widget with checkbox to disable.

    Arguments:
        type: can be "int" or "double".
        none_string: value to use when numeric entry is disabled; i.e., widget
            returns this as value if checkbox is set, and setting value to this
            will make the checkbox checked.
        note_label: text to show next to checkbox; defaults to none_string (in
            title case).
    """

    valueChanged = QtCore.Signal()

    def __init__(
        self,
        type="int",
        none_string="none",
        none_label: Optional[Text] = None,
        decimals: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super(OptionalSpinWidget, self).__init__(*args, **kwargs)

        self.none_string = none_string

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.spin_widget = (
            QtWidgets.QDoubleSpinBox() if type == "double" else QtWidgets.QSpinBox()
        )

        if type == "double" and decimals is not None:
            self.spin_widget.setDecimals(decimals)

        none_label = none_label if none_label is not None else self.none_string.title()
        self.check_widget = QtWidgets.QCheckBox(none_label)

        self.spin_widget.valueChanged.connect(self.updateState)
        self.check_widget.stateChanged.connect(self.updateState)

        layout.addWidget(self.spin_widget)
        layout.addWidget(self.check_widget)

        self.setLayout(layout)

    def updateState(self, valueChanged=True):
        self.spin_widget.setDisabled(self.check_widget.isChecked())
        if valueChanged:
            self.valueChanged.emit()

    def isNoneVal(self, val) -> bool:
        if val is None:
            return True
        if hasattr(val, "lower"):
            if self.none_string.lower() in ("", "none"):
                if val.lower() in ("none"):
                    return True
            elif self.none_string.lower() in ("auto"):
                if val.lower() in ("auto"):
                    return True
        return False

    @property
    def noneVal(self):
        if self.none_string.lower() in ("", "none"):
            return None
        else:
            return self.none_string

    def setToNone(self):
        self.setValue(self.noneVal)

    def value(self):
        if self.check_widget.isChecked():
            return self.noneVal
        return self.spin_widget.value()

    def setValue(self, val):
        is_none = self.isNoneVal(val)
        self.check_widget.setChecked(is_none)
        if not is_none:
            self.spin_widget.setValue(val)
        self.updateState(valueChanged=False)

    def setRange(self, min, max):
        self.spin_widget.setRange(min, max)


class FieldComboWidget(QtWidgets.QComboBox):
    """
    Custom ComboBox-style widget with method to easily add set of options.

    Arguments:
        result_as_idx: If True, then set/get for value will use idx of option
            rather than string.
        add_blank_option: If True, then blank ("") option will be added at
            beginning of list (which will return "" as val instead of idx if
            result_as_idx is True).
    """

    def __init__(
        self,
        result_as_idx: bool = False,
        add_blank_option: bool = False,
        *args,
        **kwargs,
    ):
        super(FieldComboWidget, self).__init__(*args, **kwargs)
        self.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.setMinimumContentsLength(3)
        self.result_as_idx = result_as_idx
        self.add_blank_option = add_blank_option
        self.options_list = []

    def set_options(self, options_list: List[Text], select_item: Optional[Text] = None):
        """
        Sets list of menu options.

        Args:
            options_list: List of items (strings) to show in menu.
            select_item: Item to select initially.

        Returns:
            None.
        """
        self.clear()
        self.options_list = options_list

        if self.add_blank_option:
            self.addItem("")
        for item in options_list:
            if item == "---":
                self.insertSeparator(self.count())
            else:
                self.addItem(item)
        if select_item is not None:
            self.setValue(select_item)

    def value(self):
        if self.result_as_idx:
            val = self.currentIndex()
            if self.add_blank_option:
                val -= 1
        else:
            val = self.currentText()

        return val

    def setValue(self, val):
        if type(val) == int and val < len(self.options_list) and self.result_as_idx:
            val = self.options_list[val]
        super(FieldComboWidget, self).setCurrentText(str(val))


class StringListWidget(QtWidgets.QLineEdit):
    """
    Free-form text field which converts value to/from list.

    Arguments:
        delim: the list delimiter to use; note that list values won't be
            trimmed, so "," and "item a, item b" will result in
            ["item a", " item b"].
    """

    def __init__(self, delim=" ", *args, **kwargs):
        super(StringListWidget, self).__init__(*args, **kwargs)
        self.delim = delim

    def setValue(self, val):
        # Check if we have a list (not *sequence* since this would include str)
        if isinstance(val, list):
            val = self.delim.join(val)
        self.setText(str(val))

    def value(self):
        return self.text().split(self.delim)


class TextOrListWidget(QtWidgets.QWidget):
    """
    Widget with free-form text field or drop-down menu.

    The "text" or "list" mode can be set using `setMode` method.

    This widget is useful when we want a drop-down menu but need to fall back
    to a text field when (e.g.) the current value isn't in list.

    Arguments:
        result_as_idx: If True, then set/get for value will use idx of option
            rather than string.
        add_blank_option: If True, then blank ("") option will be added at
            beginning of list (which will return "" as val instead of idx if
            result_as_idx is True).
    """

    valueChanged = QtCore.Signal()

    def __init__(self, result_as_idx=False, add_blank_option=False, *args, **kwargs):
        super(TextOrListWidget, self).__init__(*args, **kwargs)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.text_widget = QtWidgets.QLineEdit()
        self.list_widget = FieldComboWidget(
            result_as_idx=result_as_idx, add_blank_option=add_blank_option,
        )

        layout.addWidget(self.text_widget)
        layout.addWidget(self.list_widget)

        self.setLayout(layout)

        self.setMode("text")

        self.list_widget.currentIndexChanged.connect(self.emitValueChanged)
        self.text_widget.textChanged.connect(self.emitValueChanged)

    def emitValueChanged(self, *args):
        self.valueChanged.emit()

    def setMode(self, mode):
        self.mode = mode
        self.text_widget.setVisible(self.mode == "text")
        self.list_widget.setVisible(self.mode == "list")

    def value(self):
        if self.mode == "text":
            return self.text_widget.text()
        else:
            return self.list_widget.value()

    def setValue(self, val):
        self.text_widget.setText(str(val))
        self.list_widget.setValue(val)

    def set_options(self, *args, **kwargs):
        self.setMode("list")
        self.list_widget.set_options(*args, **kwargs)


class ResizingStackedWidget(QtWidgets.QStackedWidget):
    """
    QStackedWidget that updates its sizeHint and minimumSizeHint as needed.
    """

    def __init__(self, *args, **kwargs):
        super(ResizingStackedWidget, self).__init__(*args, **kwargs)

    def sizeHint(self):
        """Qt method."""
        return self.currentWidget().sizeHint()

    def minimumSizeHint(self):
        """Qt method."""
        return self.currentWidget().minimumSizeHint()


"""[summary]

TODO: From/to traces
"""
import numpy as np
import xarray as xr
import pandas as pd
from collections import UserDict
from typing import Optional, List, Dict, Any, Union

class Events(UserDict):

    def __init__(self, data: Optional[Dict[str, List[float]]] = None, categories: Optional[Dict[str, str]] = None,
                 add_names_from_categories: bool = True):
        """[summary]

        Args:
            data: dict or Events
            categories (dict[str: str]): name, category mapping

        """
        if data is None:
            data = dict()

        super().__init__(data)

        for key, val in self.items():
            val = np.array(val)
            if val.ndim==1:
                val = val[:, np.newaxis]
            if val.shape[1]==1:
                val = np.concatenate((val, val), axis=1)

            self.data[key] = val
        self.categories = self._infer_categories()

        # drop nan
        self._drop_nan()

        # preserve cats from input
        if hasattr(data, 'categories'):
            for name, cat in data.categories.items():
                if name in self:  # update only existing keys
                    self.categories[name] = cat

        # update cats from arg
        if categories is not None:
            for name, cat in categories.items():
                if name in self:  # update only existing keys
                    self.categories[name] = cat
                elif add_names_from_categories:
                    self.add_name(name=name, category=cat)

    @classmethod
    def from_df(cls, df, possible_event_names=[]):
        return cls.from_lists(df.name.values,
                              df.start_seconds.values.astype(np.float),
                              df.stop_seconds.values.astype(np.float),
                              possible_event_names)

    @classmethod
    def from_lists(cls, names, start_seconds, stop_seconds, possible_event_names=[]):
        unique_names = list(set(names))
        unique_names.extend(possible_event_names)
        dct = {name: [] for name in unique_names}

        for name, start_second, stop_second in zip(names, start_seconds, stop_seconds):
            dct[name].append([start_second, stop_second])

        return cls(dct)

    @classmethod
    def from_dataset(cls, ds):
        start_seconds = np.array(ds.event_times.sel(event_time='start_seconds').data)
        stop_seconds = np.array(ds.event_times.sel(event_time='stop_seconds').data)
        names = np.array(ds.event_names.data)
        if 'possible_event_names' in ds.attrs:
            possible_event_names = ds.attrs['possible_event_names']
        elif 'possible_event_names' in ds.event_names.attrs:
            possible_event_names = ds.event_names.attrs['possible_event_names']
        else:
            possible_event_names = []

        out = cls.from_lists(names, start_seconds, stop_seconds, possible_event_names)
        if 'event_categories' in ds:
            cats = {str(cat.event_types.data): str(cat.event_categories.data) for cat in ds.event_categories}
            out = cls(out, categories=cats)
        return out

    def update(self, new_dict):
        """Add all items in new_dict to self, overwrite existing items.
        Same as python's dict.update but also keeps track of categories.

        Args:
            new_dict ([type]): [description]
        """
        super().update(new_dict)
        if hasattr(self, 'categories') and hasattr(new_dict, 'categories'):
            self.categories.update(new_dict.categories)

    def _init_df(self):
        return pd.DataFrame(columns=['name', 'start_seconds', 'stop_seconds'])

    def _append_row(self, df, name, start_seconds, stop_seconds=None):
        if stop_seconds is None:
            stop_seconds = start_seconds
        new_row = pd.DataFrame(np.array([name, start_seconds, stop_seconds])[np.newaxis,:],
                                columns=df.columns)
        return df.append(new_row, ignore_index=True)

    def to_df(self, preserve_empty: bool = True):
        """Convert to pandas.DataFeame

        Args:
            preserve_empty (bool, optional):
                In keeping with the convention that events have identical start and stop times and segments do not,
                empty events are coded with np.nan as both start and stop and
                empty segments are coded as np.nan as start and 0 as stop.
                `from_df()` will obey this convention - if both start and stop are np.nan,
                the name will be a segment,
                if only the start is np.nan (the stop does not matter), the name will be an event
                Defaults to True.

        Returns:
            pandas.DataFrame: with columns name, start_seconds, stop_seconds, one row per event.
        """
        df = self._init_df()
        for name in self.names:
            for start_second, stop_second in zip(self.start_seconds(name), self.stop_seconds(name)):
                df = self._append_row(df, name, start_second, stop_second)
        if preserve_empty:  # ensure we keep events without annotations
            for name, cat in zip(self.names, self.categories.values()):
                if name not in df.name.values:
                    stop_seconds = np.nan if cat == 'event' else 0  # (np.nan, np.nan) -> empty events, (np.nan, some number) -> empty segments
                    df = self._append_row(df, name, start_seconds=np.nan, stop_seconds=stop_seconds)
        # make sure start and stop seconds are numeric
        df['start_seconds'] = pd.to_numeric(df['start_seconds'], errors='coerce')
        df['stop_seconds'] = pd.to_numeric(df['stop_seconds'], errors='coerce')
        return df

    def to_lists(self, preserve_empty: bool = True):
        """[summary]

        Args:
            preserve_empty (bool, optional):
                In keeping with the convention that events have identical start and stop times and segments do not,
                empty events are coded with np.nan as both start and stop and
                empty segments are coded as np.nan as start and 0 as stop.
                `from_df()` will obey this convention - if both start and stop are np.nan,
                the name will be a segment,
                if only the start is np.nan (the stop does not matter), the name will be an event
                Defaults to True.

        Returns:
            Tuple[List[str], List[float], List[float]: with names, start_seconds, stop_seconds.
        """
        df = self.to_df(preserve_empty=preserve_empty)
        names = df.name.values
        start_seconds = df.start_seconds.values.astype(np.float)
        stop_seconds = df.stop_seconds.values.astype(np.float)
        return names, start_seconds, stop_seconds

    def to_dataset(self):
        names, start_seconds, stop_seconds = self.to_lists()

        da_names = xr.DataArray(name='event_names', data=np.array(names, dtype='U128'), dims=['index',])
        da_times = xr.DataArray(name='event_times', data=np.array([start_seconds, stop_seconds]).T, dims=['index','event_time'], coords={'event_time': ['start_seconds', 'stop_seconds']})

        ds = xr.Dataset({da.name: da for da in [da_names, da_times]})
        ds.attrs['time_units'] = 'seconds'
        ds.attrs['possible_event_names'] = self.names  # ensure that we preserve even names w/o events that get lost in to_df
        return ds

    def add_name(self, name: str, category='segment', times=None, overwrite: bool = False, append: bool = False, sort_after_append: bool = False):
        """[summary]

        Args:
            name (str): Name of the segment/event.
            category (str, optional): Song type category ('segment' or 'event'). Defaults to 'segment'.
            times (np.array, optional): [N,2] array of floats with start (index 0) and end (index 1) of the annotations.
                                        Defaults to None.
            overwrite (bool, optional): Replace times and category if name exists. Defaults to False.
            append (bool, optional): Append times if name exists. Defaults to False.
            sort_after_append (bool, optional): Sort times by start_seconds. Defaults to False.
        """
        if times is None:
            times = np.zeros((0,2))

        if name not in self or (name in self and overwrite):
            self.update({name: times})
            self.categories[name] = category
        elif name in self and append:
            self[name] = np.append(self[name], times, axis=0)
            if sort_after_append:
                self[name].sort(axis=0)

    def delete_name(self, name):
        """Delete all annotations with that name."""
        if name in self:
            del self[name]
        if name in self.categories:
            del self.categories[name]

    def add_time(self, name: str, start_seconds: float, stop_seconds: float = None,
                 add_new_name: bool = True, category: Optional[str] = None):
        """Add a new segment/event.

        Args:
            name (str): name of the segment/event.
            start_seconds (float): start of the segment/event
            stop_seconds (float, optional): end of the segment/event (for events, should equal start). Defaults to None (use start_seconds).
            add_new_name (bool, optional): Add new song type if name does not exist yet. Defaults to True.
            category (str, optional): Manually specify category here.
        """
        if stop_seconds is None:
            stop_seconds = start_seconds

        if name not  in self and add_new_name:
            if category is None:
                category = 'event' if stop_seconds==start_seconds else 'segment'
            self.add_name(name, category=category)

        self[name] = np.insert(self[name],
                               len(self[name]),
                               sorted([start_seconds, stop_seconds]),
                               axis=0)

    def move_time(self, name, old_time, new_time):
        """[summary]

        Args:
            name ([type]): [description]
            old_time ([type]): [description]
            new_time ([type]): [description]
        """
        self[name][self[name]==old_time] = new_time

    def delete_time(self, time: float, name: Optional[str] = None, tol: float = 0, min_time: Optional[float] = None, max_time=None):
        """[summary]

        Args:
            name ([type], optional): [description]. Defaults to None.
            time ([type], optional): [description]. Defaults to None.
            tol (int, optional): [description]. Defaults to 0.
            min_time ([type], optional): [description]. Defaults to None.
            max_time ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        # TODO:
        # [ ] if hover - relevant for overlapping segments - delete that event (detect hover by checking self.mouseHovering for all LinearRegionItems in view)
        # [x] otherwise - provide min/max time and delete neareast event in view
        # [x] delete by matching both start and stop
        # [ ] refactor - have basic `delete_time(name, time)`, and then `delete_nearest(name, time, min_time, max_time)``
        if name is None:
            nearest_starts = dict()
            nearest_stops = dict()
            for name in self.keys():
                within_range_indices = self.select_range(name, min_time, max_time, strict=False)
                if len(within_range_indices):
                    nearest_starts[name] = self._find_nearest(self.start_seconds(name)[within_range_indices], time)
                    nearest_stops[name] = self._find_nearest(self.stop_seconds(name)[within_range_indices], time)

            if len(nearest_starts):
                nearest_starts_times = list(nearest_starts.values())
                nearest_stops_times = list(nearest_stops.values())
                nearest_names = list(nearest_starts.keys())
                distance_start = np.abs(time - np.array(nearest_starts_times))
                distance_stop = np.abs(time - np.array(nearest_stops_times))
                if np.min(distance_start) < np.min(distance_stop):
                    name = nearest_names[np.nanargmin(distance_start)]
                else:
                    name = nearest_names[np.nanargmin(distance_stop)]
            else:
                return None, []

        within_range_indices = self.select_range(name, min_time, max_time, strict=False)
        if len(within_range_indices):
            nearest_start = self._find_nearest(self.start_seconds(name)[within_range_indices], time)
            nearest_stop = self._find_nearest(self.stop_seconds(name)[within_range_indices], time)
        else:
            nearest_start = None
            nearest_stop = None

        if nearest_start is None and nearest_stop is None:
            return None, []

        if np.abs(nearest_start - time) < np.abs(nearest_stop - time):
            index = np.where(self.start_seconds(name) == nearest_start)[0][0]
            nearest_is_start = True
        else:
            index = np.where(self.stop_seconds(name) == nearest_stop)[0][0]
            nearest_is_start = False

        if self.categories[name] == 'segment':
            if nearest_is_start:
                matching_stop = self.stop_seconds(name)[index]
                event_at_time = matching_stop > time
            else:
                matching_start = self.start_seconds(name)[index]
                event_at_time = matching_start < time
        elif self.categories[name] == 'event':
            if nearest_is_start:
                event_at_time = np.abs(time - nearest_start) < tol
            else:
                event_at_time = np.abs(time - nearest_stop) < tol
        else:
            event_at_time = False

        if event_at_time:
            deleted_time = self[name][index, :]
            deleted_name = name
            self[name] = np.delete(self[name], index, axis=0)
        else:
            deleted_time = []
            deleted_name = None

        return deleted_name, deleted_time

    def select_range(self, name: str, t0: Optional[float] = None, t1: Optional[float] = None, strict: bool = True):
        """Get indices of events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name (str): [description]
            t0 (float, optional): [description]
            t1 (float, optional): [description]
            strict (bool, optional): if true, only matches events that start AND stop within the range,
                                     if false, matches events that start OR stop within the range

        Returns:
            List[uint]: List of indices of events within the range
        """
        if t0 is None:
            t0 = 0
        if t1 is None:
            t1 = np.inf

        if strict:
            within_range = np.logical_and(self.start_seconds(name)>t0, self.stop_seconds(name)<t1)
        else:
            starts_in_range = np.logical_and(self.start_seconds(name)>t0, self.start_seconds(name)<t1)
            stops_in_range = np.logical_and(self.stop_seconds(name)>t0, self.stop_seconds(name)<t1)
            within_range = np.logical_or(starts_in_range, stops_in_range)
        within_range_indices = np.where(within_range)[0]
        return within_range_indices

    def filter_range(self, name, t0, t1, strict: bool = False):
        """Returns events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name ([type]): [description]
            t0 ([type]): [description]
            t1 ([type]): [description]
            strict (bool): if true, only matches events that start AND stop within the range,
                           if false, matches events that start OR stop within the range
        Returns:
            List[float]: [N, 2] list of start_seconds and stop_seconds in the range
        """
        indices = self.select_range(name, t0, t1, strict)
        return self[name][indices, :]

    def delete_range(self, name, t0, t1, strict: bool = True):
        """Deletes events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name ([type]): [description]
            t0 ([type]): [description]
            t1 ([type]): [description]
            strict (bool): if true, only matches events that start AND stop within the range,
                           if false, matches events that start OR stop within the range
        Returns:
            int: number of deleted events
        """
        indices = self.select_range(name, t0, t1, strict)
        self[name] = np.delete(self[name], indices, axis=0)
        return len(indices)

    def find_next(self, t: float, names: Optional[List[str]] = None):
        """Find event starting after `t` of type in names.

        Args:
            t (float): Current time in seconds.
            names (Optional[List[str]], optional):
                List of event names to find next time in.
                Defaults to None (search in all events of any name).

        Returns:
            float or None: time of nearest next event. None if there is no next.
        """
        if names is None:
            names = self.names

        nxt = []
        for name in names:
            cmp = self[name][:, 0] > t
            if np.any(cmp):
                nxt.append(self[name][np.argmax(cmp), 0])
        if len(nxt):
            return np.min(nxt)


    def find_prev(self, t: float, names: Optional[List[str]] = None):
        """Find event ending before `t` of type in names

        Args:
            t (float): Current time in seconds.
            names (Optional[List[str]], optional):
                List of event names to find prev time in.
                Defaults to None (search in all events of any name).

        Returns:
            float or None: time of nearest previous event. None if there is no previous.
        """
        if names is None:
            names = self.names

        nxt = []
        for name in names:
            cmp = self[name][:, 1] < t
            if np.any(cmp):
                nxt.append(self[name][np.argmin(cmp)-1, 0])
        if len(nxt):
            return np.max(nxt)  #* self.fs_song

    def _find_nearest(self, array, value):
        if not len(array):
            return None
        else:
            idx = (np.abs(array - value)).argmin()
            return array[idx]

    def _infer_categories(self):
        categories = dict()
        for name in self.names:
            if len(self[name])==0:
                if not hasattr(self, 'categories') or name not in self.categories:
                    categories[name] = None
                elif hasattr(self, 'categories') and name in self.categories:
                    categories[name] = self.categories[name]
            else:
                first_start = self.start_seconds(name)[0]
                first_stop = self.stop_seconds(name)[0]

                if (np.isnan(first_start) and np.isnan(first_stop)) or (first_start == first_stop):
                    category = 'event'
                else:
                    category = 'segment'

                categories[name] = category

        return categories

    def _drop_nan(self):
        # remove entries with nan stop or start (but keep their name)
        for name in self.names:
            nan_events = np.logical_or(np.isnan(self.start_seconds(name)),
                                       np.isnan(self.stop_seconds(name)))
            self[name] = self[name][~nan_events]

    @property
    def names(self):
        return list(self.keys())

    def start_seconds(self, key):
        return self[key][:, 0]

    def stop_seconds(self, key):
        return self[key][:, 1]

    def duration_seconds(self, key):
        return self[key][:, 1] - self[key][:, 0]



"""Calculate metrics from behavioral data."""
import numpy as np
import scipy.signal.windows


def smooth(x, winlen: int = 201):
    """Smooth data along time dimension (axis=0) with Gaussian window.

    Args:
        x ([type]): [description]
        winlen (int, optional): duration of gaussian window in samples. std will be winlen/8. Defaults to 201 (std=25 samples).

    Returns:
        [type]: [description]
    """
    win = scipy.signal.windows.gaussian(winlen, winlen/8)
    win /= sum(win)
    x = scipy.ndimage.convolve1d(x, win, axis=0, mode='nearest')
    return x


def remove_nan(x):
    """Replace any nan value with nearest valid value."""
    mask = np.isnan(x)
    x[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), x[~mask])
    return x


def remove_multi_nan(x):
    """Replace any nan value with nearest valid value."""
    if x.ndim < 2:
        mask = np.isnan(x)
        x[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), x[~mask])
    elif x.ndim == 2:
        for x_feature in x.T:
            mask = np.isnan(x_feature)
            x_feature[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), x_feature[~mask])
    else:
        raise ValueError(f'Invalid arguments: x can only be one or two dimensional. Has {x.ndim} dimensions.')
    return x


def distance(pos1, pos2=None, set_self_to_nan: bool = False, exclude_cross_terms: bool = False):
    """Compute pairwise euclidean distances.

    Args:
        pos1 ([type]): positions of the thorax. [time, flies1, y/x]
        pos2 ([type], optional): positions of the thorax. [time, flies2, y/x], Defaults to None
        set_self_to_nan (bool, optional): set self-distances (diagonal along 2nd and 3rd dim) to nan. Defaults to False. Will be ignored if pos2 is not None.
        exclude_cross_terms (boolean, optional):, Defaults to False.

    Returns:
        [type]: distances between flies. [time, flies1, flies2]
    """
    if pos2 is not None:
        set_self_to_nan = False

    if pos2 is None:
        pos2 = pos1

    if exclude_cross_terms:
        dis = np.sqrt(np.sum((pos1 - pos2)**2, axis=-1))
    else:
        dis = np.sqrt(np.sum((pos1[:, np.newaxis, :, :] - pos2[:, :, np.newaxis, :])**2, axis=-1))

    if set_self_to_nan:
        nb_flies = pos1.shape[1]
        dis[..., range(nb_flies), range(nb_flies)] = np.nan  # set diag to np.nan
    return dis


def yx2fwdlat(pos1, pos2, dyx):
    """[summary]

    Args:
        pos1 ([type]): [description]
        pos2 ([type]): [description]
        dyx ([type]): [description]

    Returns:
        [type]: [description]
    """
    mags = np.linalg.norm(dyx, axis=2)  # acceleration magnitude
    angle_diff = angle(dyx, degrees=False) - angle(pos1, pos2, degrees=False)  # angle difference between dyx and orientation
    fwdlat = np.empty_like(dyx)    # acceleration components with reference to self orientation
    fwdlat[:, :, 0] = mags*np.cos(angle_diff)    # forward
    fwdlat[:, :, 1] = -mags*np.sin(angle_diff)    # lateral
    return fwdlat


def derivative(x, dt: float = 1, degree: int = 1, axis: int = 0):
    """Calculate derivative of x with respect to axis.

    Args:
        x ([type]): Data. At last 1D.
        dt (float, optional): Timestep. Defaults to 1.
        degree (int, optional): Number of time the gradient is taken. Defaults to 1 (velocity). degree=2 return acceleration.
        axis (int, optional): Axis along which to take the gradient. Defaults to 0.

    Returns:
        [type]: [description]
    """
    for _ in range(degree):
        x = np.gradient(x, dt, axis=axis)
    return x


def velocity(pos1, pos2: np.array = None, timestep: float = 1, ref: str = 'self'):
    """
    arg:
        pos1: position of vector's base, center of agent. [time, agent, y/x]
        pos2: position of vector's head, head of agent. [time, agent, y/x]
        timestep: time difference between data points (default = 1)
        ref: type of output velocity. 'self' for forward/lateral velocity components, 'chamber' for y/x components.
    returns:
        vels: change of variable with respect to time. [time,agent,y/x]
    """
    vels = derivative(pos1, timestep, degree=1)
    if ref == 'self':
        vels = yx2fwdlat(pos1, pos2, vels)
    return vels


def acceleration(pos1, pos2: np.array = None, timestep: float = 1, ref: str = 'self'):
    """
    arg:
        pos1: position of vector's base, center of agent. [time, flies, y/x]
        pos2: position of vector's head, head of agent. [time, flies, y/x]
        timestep: time difference between data points (default = 1)
        ref: type of output components. 'self' for forward/lateral components, 'chamber' for y/x components.
    returns:
        accs: second order change of variable with respect to time. [time,agent,y/x]
    """

    # acceleration in reference to chamber
    accs = derivative(pos1, timestep, degree=2)
    if ref == 'self':
        accs = yx2fwdlat(pos1, pos2, accs)
    return accs


def angle(pos1, pos2=None, degrees: bool = True, unwrap: bool = False):
    """Compute angle.

    Args:
        pos1 ([type]): position of vector's base, center of agent. [time, agent, y/x]
        pos2 ([type], optional): position of vector's head, head of agent. [time, agent, y/x]. If provided will compute angle between pos1 and pos2. Defaults to None.
        degrees (bool, optional): [description]. Defaults to True.
        unwrap (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: orientation of flies with respect to chamber. 0 degrees when flies look towards positive x axis, 90 degrees when vertical and looking upwards. [time, flies]
    """
    if pos2 is None:
        ang = np.arctan2(pos1[..., 0], pos1[..., 1])
    else:
        ang = np.arctan2(pos2[..., 0] - pos1[..., 0], pos2[..., 1] - pos1[..., 1])

    if unwrap:
        ang = np.unwrap(ang)

    if degrees:
        ang = ang * 180.0 / np.pi
    return ang


def relative_angle(pos1, pos2):
    """ Angle between agents. An element (k,i,j) from the output is the angle at kth sample between ith (reference head) and jth (target base).
    arg:
        pos1: positions of the thoraces for all flies. [time, flies, y/x]
        pos2: positions of the heads for all flies. [time, flies, y/x]
    returns:
        rel_angles: orientation of flies with respect to chamber. [time, flies, flies]
    """
    d0 = pos2 - pos1
    d1 = pos1[:, np.newaxis, :, :] - pos2[:, :, np.newaxis, :]  # all pairwise "distances"

    dot = d0[:, :, np.newaxis, 1]*d1[:, :, :, 1] + d0[:, :, np.newaxis, 0]*d1[:, :, :, 0]
    det = d0[:, :, np.newaxis, 1]*d1[:, :, :, 0] - d0[:, :, np.newaxis, 0]*d1[:, :, :, 1]

    rel_angles = np.arctan2(det, dot)

    return rel_angles * 180.0 / np.pi


def rot_speed(pos1, pos2, timestep: float = 1):
    """
    arg:
        pos1: position of vector's base, center of agent. [time, flies, y/x]
        pos2: position of vector's head, head of agent. [time, flies, y/x]
    returns:
        rot_speed: rotational speed. [time, flies]
    """
    unwrapped_angles = angle(pos1, pos2, unwrap=True)
    rot_speed = derivative(unwrapped_angles, timestep, degree=1)
    return rot_speed


def rot_acceleration(pos1, pos2, timestep: float = 1):
    """[summary]

    Args:
        pos1 ([type]): position of vector's base, center of agent. [time, flies, y/x]
        pos2 ([type]): position of vector's head, head of agent. [time, flies, y/x]
        timestep (float, optional): [description]. Defaults to 1.

    Returns:
        [type]: rotational acceleration. [time, flies]
    """
    unwrapped_angles = angle(pos1, pos2, unwrap=True)
    rot_accs = derivative(unwrapped_angles, timestep, degree=2)
    return rot_accs


def project_velocity(vx, vy, radians):
    """Rotate vectors around the origin (0, 0).

    Args:
        vx ([type]): vector x component.
        vy ([type]): vector y component.
        radians ([type]): angles of rotation in radians.

    Returns:
        rot_vx ([type]): vector lateral component. Left negative.
        rot_vy ([type]): vector forward component. Backwards negative
    """

    if radians.shape != vx.shape:
        radians = np.repeat(radians[..., np.newaxis], radians.shape[1], axis=2)

    rot_vx = -vx * np.cos(radians) + vy * np.sin(radians)
    rot_vy = vx * np.sin(radians) + vy * np.cos(radians)

    return rot_vx, rot_vy


def internal_angle(A: np.array, B: np.array, C: np.array, deg:bool=True, array_logic: str='tfc'):
    """Calculates internal angle (∠ABC) between three points (aka, angle between lines AB and BC). If A,B,C are lists or arrays, calculation happens element-wise.
    
    Args:
        A, B, C ([type]): position of points between which the angle is calculated.
        deg ([type]): Return angle in degrees if True, radians if False (default).

    Returns:
        angles ([type]): internal angle between lines AB and BC.
    """

    v1s = B-A # vector from point B to A
    v2s = B-C # vector from point B to C

    # reshape vector arrays to be [time, coordinates, ...]
    if A.ndim == 3:
        if array_logic == 'tfc':
            v1s = np.swapaxes(v1s,1,2)
            v2s = np.swapaxes(v2s,1,2)
        elif array_logic == 'ftc':
            v1s = np.swapaxes(v1s,1,2)
            v1s = np.swapaxes(v1s,0,2)
            v2s = np.swapaxes(v2s,1,2)
            v2s = np.swapaxes(v2s,0,2)
        elif array_logic == 'fct':
            v1s = np.swapaxes(v1s,0,2)
            v2s = np.swapaxes(v2s,0,2)
    elif A.ndim > 3:
        print('Result might not be correct, only tested for arrays with 2 or 3 dimensions. Contact Adrian for help, if required.')

    dot_v1_v2 = np.einsum('ij...,ij...->i...', v1s, v2s)
    dot_v1_v1 = np.einsum('ij...,ij...->i...', v1s, v1s)
    dot_v1_v1[dot_v1_v1 == 0] = 0.0001 # small values for dot products which happen to be zero (shouldn't happen)
    dot_v2_v2 = np.einsum('ij...,ij...->i...', v2s, v2s)
    dot_v2_v2[dot_v2_v2 == 0] = 0.0001 # small values for dot products which happen to be zero (shouldn't happen)

    angles = np.arccos(dot_v1_v2/np.sqrt(dot_v1_v1*dot_v2_v2))

    if deg:
        angles *= 180/np.pi

    return angles


import xarray_behave as xb
import logging

logging.getLogger().setLevel(logging.INFO)

recs = {'rpi9-20210409_093149': {},
        'localhost-20210617_113024': {},
        'localhost-20181120_144618': {},
        'localhost-20210624_104612': {},
        'localhost-20210628_145223': {'dat_path': 'dat', 'res_path': 'res', 'root': 'tests/data'},
        'localhost-20210629_171532': {'dat_path': 'dat', 'res_path': 'res', 'root': 'tests/data'},
        'Dmel_male': {'dat_path': 'dat', 'res_path': 'dat', 'root': 'tests/data', 'filepath_daq': 'dat/Dmel_male.wav', 'filepath_annotations': 'dat/Dmel_male_annotations.csv'},
        'Dmel_male2': {'dat_path': 'dat', 'res_path': 'dat', 'root': 'tests/data', 'filepath_daq': 'dat/Dmel_male.npz', 'filepath_annotations': 'dat/Dmel_male_annotations.csv'}}


# TODO: add asserts to ensure results conform to expected ds structure
# def test_assemble0():
#     datenames = list(recs.keys())
#     ii = 0
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble1():
#     datenames = list(recs.keys())
#     ii = 1
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble2():
#     datenames = list(recs.keys())
#     ii = 2
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble3():
#     datenames = list(recs.keys())
#     ii = 3
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble4():
#     datenames = list(recs.keys())
#     ii = 4
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)


# def test_assemble5():
#     datenames = list(recs.keys())
#     ii = 5
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble6():
#     datenames = list(recs.keys())
#     ii = 6
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)

# def test_assemble7():
#     datenames = list(recs.keys())
#     ii = 7
#     datename = datenames[ii]
#     kwargs = recs[datename]
#     ds = xb.assemble(datename, **kwargs)


if __name__ == '__main__':
    test_assemble0()
    test_assemble1()
    test_assemble2()
    test_assemble3()
    test_assemble4()
    test_assemble5()
    test_assemble6()
    test_assemble7()


"""Poses loaders

should return:
    poses: np.array[frames, flies, body_parts, x/y] - in egocentric coordinates
    poses_allo: np.array[frames, flies, body_parts, x/y] - in allocentric (frame) coordinates
    body_parts: List[str]
    first_pose_frame: int
    last_pose_frame: int
"""
import h5py
import numpy as np
import xarray as xr
import pandas as pd
import zarr
import os
from .. import io
import math
from typing import Optional
import logging
import json


def rotate_point(pos, degrees, origin=(0, 0)):
    """Rotate point.

    Args:
        pos (tuple): (x.y) position
        degrees ([type]): degree by which to rotate
        origin (tuple, optional): point (x,y) around which to rotate point. Defaults to (0, 0).

    Returns:
        tuple: (x, y) rotated
    """
    x, y = pos
    radians = degrees / 180 * np.pi
    offset_x, offset_y = origin
    adjusted_x = (x - offset_x)
    adjusted_y = (y - offset_y)
    cos_rad = math.cos(radians)
    sin_rad = math.sin(radians)
    qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
    qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y
    return qx, qy


def rotate_pose(positions, degree, origin=(0, 0)):
    """Rotate point set.

    Args:
        pos (np.ndarray): [points, x/y]
        degrees (float): degree by which to rotate
        origin (Tuple[float, float], optional): point (x,y) around which to rotate point. Defaults to (0, 0).

    Returns:
        tuple: (x, y) rotated
    """
    radians = degree / 180 * np.pi
    cos_rad = np.cos(radians)
    sin_rad = np.sin(radians)
    rot = np.array([[cos_rad, sin_rad], [-sin_rad, cos_rad]])
    positions_centered = positions - origin
    return rot.dot(positions_centered.T).T + origin


def load_skeleton(filename, body_parts):
    skeleton_file = os.path.splitext(filename)[0] + '_skeleton.csv'
    if os.path.exists(skeleton_file):
        logging.info(f'Loading skeleton from {skeleton_file}.')
        df = pd.read_csv(skeleton_file)
        skeleton = df.values
    elif len(body_parts) == 11:  # probably a fly - use default skeleton
        logging.info(f'No skeleton file ({skeleton_file}) found but detected 11 - assume this is a fly.')
        skeleton = np.array([[0, 1], [1, 8], [8, 11],  # body axis
                        [8, 2], [8, 3], [8, 4], [8, 5], [8, 6], [8, 7],  # legs
                        [8, 9], [8, 10]])  # wings
    else:
        logging.info(f'No skeleton file ({skeleton_file}) found and poses do not have 11 body parts - falling back to empty skeleton.')
        skeleton = np.zeros((0, 2), dtype=np.uint)

    return skeleton


class Poses():
    def make(self, filename: Optional[str] = None):
        pixel_size_mm = None
        if filename is None:
            filename = self.path
        poses, poses_allo, body_parts, first_pose_frame, last_pose_frame = self.load(filename)

        # load skeleton from file
        skeleton = load_skeleton(filename, body_parts)

        xr_poses = xr.DataArray(data=poses,
                                 dims=['frame_number', 'flies', 'poseparts', 'coords'],
                                 coords={'frame_number': np.arange(first_pose_frame, last_pose_frame),
                                         'poseparts': body_parts,
                                         'coords': ['y', 'x']},
                                 attrs={'description': 'coords are "allocentric" - rel. to the full frame',
                                        'type': 'poses',
                                        'spatial_units': 'pixels',
                                        'pixel_size_mm': pixel_size_mm,
                                        'loader': self.NAME,
                                        'kind': self.KIND,
                                        'path': filename,
                                        'skeleton': skeleton})

        xr_poses_allo = xr.DataArray(data=poses_allo,
                                 dims=['frame_number', 'flies', 'poseparts', 'coords'],
                                 coords={'frame_number': np.arange(first_pose_frame, last_pose_frame),
                                         'poseparts': body_parts,
                                         'coords': ['y', 'x']},
                                 attrs={'description': 'coords are "allocentric" - rel. to the full frame',
                                        'type': 'poses',
                                        'spatial_units': 'pixels',
                                        'pixel_size_mm': pixel_size_mm,
                                        'loader': self.NAME,
                                        'kind': self.KIND,
                                        'path': filename,
                                        'skeleton': skeleton})
        return xr_poses, xr_poses_allo


@io.register_provider
class Leap(Poses, io.BaseProvider):

    KIND = 'poses'
    NAME = 'leap'
    SUFFIXES = ['_poses.h5', '_poses_leap.h5']

    def load(self, filename: Optional[str] = None):
        """Load pose tracks estimated using LEAP.

        Args:
            filepath ([type]): [description]

        Returns:
            poses, poses_allo, body_parts, first_pose_frame, last_pose_frame
        """
        if filename is None:
            filename = self.path

        with h5py.File(filename, 'r') as f:
            box_offset = f['box_centers'][:]
            box_angle = f['fixed_angles'][:].astype(np.float64)
            box_size = (f['box_size'].attrs['i0'], f['box_size'].attrs['i1'])
            poses = f['positions'][:]
            nb_flies = int(np.max(f['fly_id']) + 1)
        body_parts = ['head', 'neck', 'front_left_leg', 'middle_left_leg', 'back_left_leg', 'front_right_leg',
                      'middle_right_leg', 'back_right_leg', 'thorax', 'left_wing', 'right_wing', 'tail']
        # need to offset and rotate
        last_pose_index = np.argmin(poses[:, 0, 0] > 0)
        first_pose_index = np.argmin(poses[:, 0, 0] == 0)
        box_offset = box_offset[first_pose_index:last_pose_index, ...]
        box_angle = box_angle[first_pose_index:last_pose_index, ...]
        poses = poses[first_pose_index:last_pose_index, ...]
        poses = poses.astype(np.float64)

        # transform poses back to frame coordinates
        origin = [b / 2 for b in box_size]
        poses_allo = np.zeros_like(poses)
        for cnt, ((x, y), ang, p_ego) in enumerate(zip(box_offset, box_angle, poses)):
            # tmp = [rotate_point(pt, a, origin) for pt in p_ego]  # rotate
            tmp = rotate_pose(p_ego, float(ang), origin)
            p_allo = np.stack(tmp) + np.array((x, y)) - origin  # translate
            poses_allo[cnt, ...] = p_allo

        # center poses around thorax
        thorax_idx = 8
        thorax_pos = poses[:, thorax_idx, :]
        poses = poses - thorax_pos[:, np.newaxis, :]  # subtract remaining thorax position
        origin = (0, 0)

        # rotate poses such that the head is at 0 degrees
        head_idx = 0
        head_pos = poses[:,  head_idx, :]
        head_thorax_angle = 90 + np.arctan2(head_pos[:, 0], head_pos[:, 1]) * 180 / np.pi
        for cnt, (a, p_ego) in enumerate(zip(head_thorax_angle, poses)):
            # poses[cnt, ...] = [rotate_point(pt, -a, origin) for pt in p_ego]
            poses[cnt, ...] = rotate_pose(p_ego, -a, origin)

        box_offset = box_offset.reshape((-1, nb_flies, *box_offset.shape[1:]))  # "unfold" fly dimension
        box_angle = box_angle.reshape((-1, nb_flies, *box_angle.shape[1:]))  # "unfold" fly dimension
        poses = poses.reshape((-1, nb_flies, *poses.shape[1:]))  # "unfold" fly dimension
        poses_allo = poses_allo.reshape((-1, nb_flies, *poses_allo.shape[1:]))  # "unfold" fly dimension

        first_pose_frame = int(first_pose_index / nb_flies)
        last_pose_frame = int(last_pose_index / nb_flies)

        return poses, poses_allo, body_parts, first_pose_frame, last_pose_frame


@io.register_provider
class DeepPoseKit(Poses, io.BaseProvider):

    KIND = 'poses'
    NAME = 'deepposekit'
    SUFFIXES = ['_poses_dpk.zarr']

    def load(self, filename: Optional[str] = None):
        """Load pose tracks estimated using DeepPoseKit.

        Args:
            filepath (str): name of the file produced by DeepPoseKit

        Returns:
            poses_ego (xarray.DataArray): poses in EGOcentric (centered around thorax, head aligned straight upwards), [frames, flies, bodypart, x/y]
            poses_allo (xarray.DataArray): poses in ALLOcentric (frame) coordinates, [frames, flies, bodypart, x/y]
            partnames (List[str]): list of names for each body part (is already part of the poses_ego/allo xrs)
            first_pose_frame, last_pose_frame (int): frame corresponding to first and last item in the poses_ego/allo arrays (could attach this info as xr dim)
        """
        if filename is None:
            filename = self.path

        with zarr.ZipStore(filename, mode='r') as zarr_store:
            ds = xr.open_zarr(zarr_store).load()  # the final `load()` prevents
        nb_flies = len(ds.flies)
        box_size = np.array(ds.attrs['box_size'])

        poses_allo = ds.poses + ds.box_centers - box_size/2

        first_pose_frame = int(np.argmin(np.isnan(ds.poses.data[:, 0, 0, 0]).data))
        last_pose_frame = int(np.argmin(~np.isnan(np.array(ds.poses.data[first_pose_frame:, 0, 0, 0]).data)) + first_pose_frame)
        if last_pose_frame == first_pose_frame or last_pose_frame == 0:
            last_pose_frame = ds.poses.shape[0]

        # CUT to first/last frame with poses
        poses_ego = ds.poses[first_pose_frame:last_pose_frame, ...]
        poses_allo = poses_allo[first_pose_frame:last_pose_frame, ...]

        poses_ego = poses_ego - poses_ego.sel(poseparts='thorax')  # CENTER egocentric poses around thorax

        # ROTATE egocentric poses such that the angle between head and thorax is 0 degrees (straight upwards)
        head_thorax_angle = 270 + np.arctan2(poses_ego.sel(poseparts='head', coords='y'),
                                            poses_ego.sel(poseparts='head', coords='x')) * 180 / np.pi
        for cnt, (a, p_ego) in enumerate(zip(head_thorax_angle.data, poses_ego.data)):
            for fly in range(nb_flies):
                poses_ego.data[cnt, fly, ...] = rotate_pose(p_ego[fly], -a[fly])

        return poses_ego, poses_allo, ds.poseparts, first_pose_frame, last_pose_frame


@io.register_provider
class Sleap(Poses, io.BaseProvider):

    KIND = 'poses'
    NAME = 'sleap'
    SUFFIXES = ['_poses_sleap.h5', '_sleap.h5.slp']

    def load(self, filename: Optional[str] = None):
        if filename is None:
            filename = self.path

        with h5py.File(filename, 'r') as f:

            if 'node_names' not in f:
                meta = json.loads(f['metadata'].attrs['json'])
                pose_parts = []
                for node in meta['nodes']:
                    pose_parts.append(node['name'])
            else:
                pose_parts = f['node_names'][:]

            # track_names = f['track_names']
            # track_occupancy = f['track_occupancy'][:]
            tracks = f['tracks'][:]  # flies, (x/y), bodypart, frame

        poses_allo = tracks.transpose([3, 0, 2, 1])   # -> [frames, flies, bodypart, x/y]
        poses_allo = poses_allo[..., ::-1]  # swap x/y
        poses_ego = poses_allo.copy()
        pose_parts = [pose_part.decode() for pose_part in pose_parts]  # from byte to str

        first_pose_frame = 0
        last_pose_frame = poses_allo.shape[0]

        return poses_ego, poses_allo, pose_parts, first_pose_frame, last_pose_frame



# @io.register_provider
# class CSV_poses(Poses, io.BaseProvider):

#     KIND = 'tracks'
#     NAME = 'generic csv'
#     SUFFIXES = ['_tracks.csv']

#     def load(self, filename: Optional[str] = None):
#         """Load tracker data from CSV file.

#         Head of the CSV file should look like this:
#         track	track1	track1	track1	track1	track1	track1
#         part	a	    a	    b	    b  	    c	    c
#         coord	x	    y	    x	    y	    x	    y
#         0       0.09    0.09    0.09    0.09    0.09    0.09
#         1       0.09    0.09    0.09    0.09    0.09    0.09
#         2       0.09    0.09    0.09    0.09    0.09    0.09
#         ...

#         First column contains the framenumber (does not need to start at 0),
#         remaining columns contain coordinate date for different tracks/trackparts.

#         Args:
#             filename (Optional[str], optional): Path to the CSV file. Defaults to None.

#         Returns:
#             x: np.array([frames, tracks, parts, coords (y/x)]), track_names: List[], track_parts: List[], frame_numbers: np.array[np.intp]
#         """

#         if filename is None:
#             filename = self.path
#         logging.warning('Loading tracks from CSV.')
#         df = pd.read_csv(filename, header=[0, 1, 2], index_col=0)
#         track_names = df.columns.levels[df.columns.names.index('track')].to_list()
#         track_parts = df.columns.levels[df.columns.names.index('part')].to_list()
#         track_coord = df.columns.levels[df.columns.names.index('coord')].to_list()
#         coord_order = [track_coord.index('y'), track_coord.index('x')]
#         x = np.reshape(df.values, (-1, len(track_names), len(track_parts), len(track_coord)))
#         x = x[..., coord_order]
#         frame_numbers = df.index.to_numpy().astype(np.intp)
#         # return x, track_names, track_parts, frame_numbers
#         return poses_ego, poses_allo, ds.poseparts, first_pose_frame, last_pose_frame




"""Load parameters for DLP movies

"""
import numpy as np
import pandas as pd
import h5py
from .. import io
from typing import Optional
import logging
import xarray as xr

class MovieParams():

    def make(self, filename: Optional[str] = None):
        if filename is None:
            filename = self.path

        movieparams, first_movie_frame, last_movie_frame = self.load(filename)

        xr_balltracks = xr.DataArray(data=movieparams.values,
                                    dims=['frame_number_movie', 'params_movie'],
                                    coords={'frame_number_movie': np.arange(first_movie_frame, last_movie_frame, dtype=np.intp),
                                            'params_movie': list(movieparams.columns)},
                                    attrs={'loader': self.NAME,
                                           'kind': self.KIND,
                                           'path': filename,})
        return xr_balltracks


@io.register_provider
class DLP_h5log(io.BaseProvider, MovieParams):

    KIND = 'movieparams'
    NAME = 'h5'
    SUFFIXES = ['_dlp.h5']

    def load(self, filename: Optional[str] = None):
        """Load tracker data.

        Data are stored as nested dict:
        - first level is the name of the DLP_runner, second level is the logged parameter
        - and one single-level key (systemtime) with the system time for each frame

        Args:
            filename (Optional[str], optional): Defaults to None.

        Returns:
            dict: dict with keys {DLPRUNNER_PARAM: values}
            int, int: first_movie_frame, last_movie_frame
        """
        if filename is None:
            filename = self.path

        movieparams = dict()
        with h5py.File(filename, mode='r') as f:
            keys = list(f.keys())
            keys.remove('systemtime')

            movieparams['systemtime'] = f['systemtime'][:]
            for key in keys:
                val = f[key]
                for k, v  in val.items():
                    movieparams[f'{key}_{k}'] = v[:]

        first_movie_frame = int(0)
        last_movie_frame = int(np.median([len(movieparams[k]) for k in movieparams.keys()]))
        movieparams = pd.DataFrame(movieparams)
        return movieparams, first_movie_frame, last_movie_frame


@io.register_provider
class DLP_params(io.BaseProvider, MovieParams):

    KIND = 'movieparams'
    NAME = 'npz'
    SUFFIXES = ['_movieparams.npz']

    def load(self, filename: Optional[str] = None):
        """Load DLP movieparams from stimulus file.

        Args:
            filename (Optional[str], optional): Defaults to None.

        Returns:
            dict: dict with keys {DLPRUNNER_PARAM: values}
            int, int: first_movie_frame, last_movie_frame
        """
        if filename is None:
            filename = self.path

        movieparams = np.load(filename)
        movieparams = pd.DataFrame(dict(movieparams))
        first_movie_frame = int(0)
        last_movie_frame = int(np.median([len(movieparams[k]) for k in movieparams.keys()]))
        return movieparams, first_movie_frame, last_movie_frame



from setuptools import setup, find_packages
import codecs
import re
import os

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='xarray_behave',
      version=find_version("src/xarray_behave/__init__.py"),
      description='xarray_behave',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://github.com/janclemenslab/xarray-behave',
      author='Jan Clemens',
      author_email='clemensjan@googlemail.com',
      license='MIT',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      package_data={'xarray_behave': ['gui/forms/*', 'gui/icon.png']},
      python_requires='>=3.6',
      install_requires=['numpy', 'scipy', 'pandas', 'xarray', 'h5py>=2.9', 'zarr', 'flammkuchen',
                        'dask', 'toolz', 'samplestamps', 'pysoundfile', 'opencv-python-headless'],
      extras_require={'gui': ['pyside2', 'pyqtgraph>0.12.2', 'pyvideoreader', 'colorcet',
                              'sounddevice', 'scikit-image', 'opencv-python-headless', 'pyyaml',
                              'peakutils', 'defopt']},
      include_package_data=True,
      zip_safe=False,
      entry_points={'console_scripts': ['xb=xarray_behave.gui.app:cli']}
     )
